/********************************************************************/
/********************************************************************/
/**                                                                **/
/**                  BIBLIOTHEQUE  MATHEMATIQUE                    **/
/**                     (deuxieme partie)                          **/
/**                                                                **/
/**                     Copyright Babe Cool                        **/
/**                                                                **/
/********************************************************************/
/********************************************************************/
/* $Id: bibli2.c,v 1.922 1997/04/06 02:52:40 karim Exp karim $ */
#include "genpari.h"

/********************************************************************/
/**                                                                **/
/**                  POLYNOMIAL INTERPOLATION                      **/
/**                                                                **/
/********************************************************************/

GEN
polint(GEN xa, GEN ya, GEN x, GEN *dy)
{
  long av,tetpil,i,m,n;
  long ns=1, tx=typ(xa), ty=typ(ya), lx=lg(xa), ly=lg(ya);
  GEN den,dif,dift,ho,hp,w,y,c,d;
  GEN *gptr[2];

  if (! is_vec_t(tx) || ! is_vec_t(ty))
    err(talker,"not vectors in polinterpolate");
  if (lx != ly)
    err(talker,"different lengths in polinterpolate");
  n=lx-1; if (n<=1) { y=gcopy((GEN)ya[1]); *dy=gcopy(y); return y; }
  av=avma; c=cgetg(ly,ty); d=cgetg(ly,ty);
  dif=gabs(gsub(x,(GEN)xa[1]),MEDDEFAULTPREC);
  for (i=1; i<=n; i++)
  {
    dift = gabs(gsub(x,(GEN)xa[i]), MEDDEFAULTPREC);
    if (gcmp(dift,dif)<0) { ns=i; dif=dift; }
    c[i]=ya[i]; d[i]=ya[i];
  }
  y=(GEN)ya[ns--];
  tetpil=avma;
  for (m=1; m<n; m++)
  {
    for (i=1; i<=n-m; i++)
    {
      ho=gsub((GEN)xa[i],x); hp=gsub((GEN)xa[i+m],x);
      den=gsub(ho,hp);
      if (gcmp0(den)) err(talker,"two abcissas are equal in polint");
      w=gsub((GEN)c[i+1],(GEN)d[i]); den=gdiv(w,den);
      d[i]=lmul(hp,den); c[i]=lmul(ho,den);
    }
    *dy= (2*ns < (n-m))? (GEN)c[ns+1]: (GEN)d[ns--];
    tetpil=avma; y=gadd(y,*dy);
  }
  *dy=gcopy(*dy); gptr[0]=&y; gptr[1]=dy;
  gerepilemanysp(av,tetpil,gptr,2);
  return y;
}

/********************************************************************/
/**                                                                **/
/**                NUMERICAL INTEGRATION (Romberg)                 **/
/**                                                                **/
/********************************************************************/
GEN
intnum0(entree *ep, GEN a, GEN b, char *ch, long flag, long prec)
{
  switch(flag)
  {
    case 0: return qromb(ep,a,b,ch,prec);
    case 1: return rombint(ep,a,b,ch,prec);
    case 2: return qromi(ep,a,b,ch,prec);
    case 3: return qromo(ep,a,b,ch,prec);
    default: err(flagerr);
  }
  return NULL; /* not reached */
}

#define JMAX 25
#define JMAXP JMAX+3
#define KLOC 5

GEN
qromb(entree *ep, GEN a, GEN b, char *ch, long prec)
{
  GEN ss,dss,s,h,q1,p1,p2,p3,p4,p,qlint,del,sz,x,sum;
  long av,tetpil,j,j1,j2,lim,l,it,tnm,sig;

  av=avma; l=prec;
  if (typ(a)!=t_REAL) { p=cgetr(prec); gaffect(a,p); a=p; }
  if (typ(b)!=t_REAL) { p=cgetr(prec); gaffect(b,p); b=p; }
  qlint=subrr(b,a); sig=signe(qlint);
  if (!sig) { avma=av; return gzero; }

  if (sig<0) { setsigne(qlint,1); s=a; a=b; b=s; }
  newvalue(ep,cgetr(prec)); q1=(GEN)ep->value;
  p3=cgetg(KLOC+1,t_VEC); p4=cgetg(KLOC+1,t_VEC);
  s=cgetg(JMAXP,t_VEC); h=cgetg(JMAXP,t_VEC);
  affsr(1,(GEN)(h[1]=lgetr(l)));

  gaffect(a,q1); p1=lisexpr(ch);
  gaffect(b,q1); p2=lisexpr(ch);
  s[1]=lmul2n(gmul(qlint,gadd(p1,p2)),-1);
  
  it=1; sz=gmul(gzero,(GEN)s[1]);
  s[2]=s[1]; h[2]=lshiftr((GEN)h[1],-2);
  for (j=2; j<=JMAX; j++)
  {
    tnm=it; del=divrs(qlint,tnm); x=addrr(a,shiftr(del,-1));
    for (sum=sz,j1=1; j1<=it; j1++,x=addrr(x,del))
    {
      affrr(x,q1); p1=lisexpr(ch); sum=gadd(sum,p1);
    }
    it*=2;
    s[j]=lmul2n(gadd((GEN)s[j-1],gmul(sum,del)),-1);

    if (j>=KLOC)
    {
      for (j1=1; j1<=KLOC; j1++)
      {
	p3[j1]=s[j1+j-KLOC]; p4[j1]=h[j1+j-KLOC];
      }
      tetpil=avma; ss=polint(p4,p3,gzero,&dss);
      j1=gexpo(ss); j2=gexpo(dss); lim=((prec-2)<<TWOPOTBITS_IN_LONG)-j-5;
      if (j1-j2 > lim || (j1 < -lim && j2<j1-1))
      {
	if (gcmp0(gimag(ss))) ss=greal(ss);
	tetpil=avma; killvalue(ep);
	return gerepile(av,tetpil,gmulsg(sig,ss));
      }
    }
    s[j+1]=s[j]; h[j+1]=lshiftr((GEN)h[j],-2);
  }
  err(intger2); return NULL; /* not reached */
}

GEN
qromo(entree *ep, GEN a, GEN b, char *ch, long prec)
{
  GEN ss,dss,s,h,q1,sz,p1,p3,p4,p,qlint,del,ddel,x,sum;
  long av,tetpil,j,j1,j2,lim,l,it,tnm,sig;

  av=avma; l=prec;
  if (typ(a)!=t_REAL) { p=cgetr(prec); gaffect(a,p); a=p; }
  if (typ(b)!=t_REAL) { p=cgetr(prec); gaffect(b,p); b=p; }
  qlint=subrr(b,a); sig=signe(qlint);
  if (!sig) { avma=av; return gzero; }

  if (sig<0) { setsigne(qlint,1); s=a; a=b; b=s; }
  newvalue(ep,cgetr(prec)); q1=(GEN)ep->value;
  p3=cgetg(KLOC+1,t_VEC); p4=cgetg(KLOC+1,t_VEC);
  s=cgetg(JMAXP,t_VEC); h=cgetg(JMAXP,t_VEC);
  affsr(1,(GEN)(h[1]=lgetr(l)));

  affrr(shiftr(addrr(a,b),-1),q1); p1=lisexpr(ch);
  s[1]=lmul(qlint,p1);
  
  it=1; sz=gmul(gzero,(GEN)s[1]);
  s[2]=s[1]; h[2]=ldivrs((GEN)h[1],9);
  for (j=2; j<=JMAX; j++)
  {
    tnm=it; del=divrs(qlint,3*tnm); ddel=shiftr(del,1);
    x=addrr(a,shiftr(del,-1)); sum=sz;
    for (j1=1; j1<=it; j1++)
    {
      affrr(x,q1); p1=lisexpr(ch); sum=gadd(sum,p1); x=addrr(x,ddel);
      affrr(x,q1); p1=lisexpr(ch); sum=gadd(sum,p1); x=addrr(x,del);
    }
    it*=3;
    s[j]=ladd(gdivgs((GEN)s[j-1],3),gmul(sum,del));

    if (j>=KLOC)
    {
      for (j1=1; j1<=KLOC; j1++)
      {
	p3[j1]=s[j1+j-KLOC]; p4[j1]=h[j1+j-KLOC];
      }
      tetpil=avma; ss=polint(p4,p3,gzero,&dss);
      j1=gexpo(ss); j2=gexpo(dss); lim=((prec-2)<<TWOPOTBITS_IN_LONG)-(3*j/2)-5;
      if ( j1-j2 > lim || (j1 < -lim && j2<j1-1))
      {
	if (gcmp0(gimag(ss))) ss=greal(ss);
	tetpil=avma; killvalue(ep);
	return gerepile(av,tetpil,gmulsg(sig,ss));
      }
    }
    s[j+1]=s[j]; h[j+1]=ldivrs((GEN)h[j],9);
  }
  err(intger2); return NULL; /* not reached */
}

#undef JMAX
#undef JMAXP
#define JMAX 16
#define JMAXP JMAX+3

GEN
qromi(entree *ep, GEN a, GEN b, char *ch, long prec)
{
  GEN ss,dss,s,h,q1,sz,p1,p3,p4,p,qlint,del,ddel,x,sum;
  long av,tetpil,j,j1,j2,lim,l,it,tnm,sig;

  av=avma; l=prec;
  p=cgetr(prec); gaffect(ginv(a),p); a=p;
  p=cgetr(prec); gaffect(ginv(b),p); b=p;
  qlint=subrr(b,a); sig= -signe(qlint);
  if (!sig) { avma=av; return gzero; }
  if (sig>0) { setsigne(qlint,1); s=a; a=b; b=s; }

  newvalue(ep,cgetr(prec)); q1=(GEN)ep->value;
  p3=cgetg(KLOC+1,t_VEC); p4=cgetg(KLOC+1,t_VEC);
  s=cgetg(JMAXP,t_VEC); h=cgetg(JMAXP,t_VEC);
  affsr(1,(GEN)(h[1]=lgetr(l))); x=divsr(2,addrr(a,b));
  affrr(x,q1);
  p1=gmul(lisexpr(ch),mulrr(x,x)); s[1]=lmul(qlint,p1); it=1;
  sz=gmul(gzero,(GEN)s[1]); s[2]=s[1]; h[2]=ldivrs((GEN)h[1],9);
  for (j=2; j<=JMAX; j++)
  {
    tnm=it; del=divrs(qlint,3*tnm); ddel=shiftr(del,1);
    x=addrr(a,shiftr(del,-1)); sum=sz;
    for (j1=1; j1<=it; j1++)
    {
      divsrz(1,x,q1); p1=gmul(lisexpr(ch),mulrr(q1,q1));
      sum=gadd(sum,p1); x=addrr(x,ddel);
      divsrz(1,x,q1); p1=gmul(lisexpr(ch),mulrr(q1,q1));
      sum=gadd(sum,p1); x=addrr(x,del);
    }
    it*=3;
    s[j]=ladd(gdivgs((GEN)s[j-1],3),gmul(sum,del));
    if (j>=KLOC)
    {
      for (j1=1; j1<=KLOC; j1++)
      {
	p3[j1]=s[j1+j-KLOC]; p4[j1]=h[j1+j-KLOC];
      }
      tetpil=avma; ss=polint(p4,p3,gzero,&dss);
      j1=gexpo(ss); j2=gexpo(dss); lim=((prec-2)<<TWOPOTBITS_IN_LONG)-(3*j/2)-5;
      if (j1-j2 > lim || (j1 < -lim && j2 < j1-1))
      {
	if (gcmp0(gimag(ss))) ss=greal(ss);
	tetpil=avma; killvalue(ep);
	return gerepile(av,tetpil,gmulsg(sig,ss));
      }
    }
    s[j+1]=s[j]; h[j+1]=ldivrs((GEN)h[j],9);
  }
  err(intger2); return NULL; /* not reached */
}

GEN
rombint(entree *ep, GEN a, GEN b, char *ch, long prec)
{
  GEN aa,bb,mun,p1,p2,p3;
  long l,av,tetpil;

  l=gcmp(b,a); if (!l) return gzero;
  if (l<0) { bb=a; aa=b; } else { bb=b; aa=a; }
  av=avma; mun = negi(gun);
  if (gcmpgs(bb,100)>=0)
  {
    if (gcmpgs(aa,1)>=0) return qromi(ep,a,b,ch,prec);
    p1=qromi(ep,gun,bb,ch,prec);
    if (gcmpgs(aa,-100)>=0)
    {
      p2=qromo(ep,aa,gun,ch,prec); tetpil=avma;
      return gerepile(av,tetpil,gmulsg(l,gadd(p1,p2)));
    }
    p2=qromo(ep,mun,gun,ch,prec); p3=gadd(p2,qromi(ep,aa,mun,ch,prec));
    tetpil=avma; return gerepile(av,tetpil,gmulsg(l,gadd(p1,p3)));
  }
  if (gcmpgs(aa,-100)>=0) return qromo(ep,a,b,ch,prec);
  if (gcmpgs(bb,-1)>=0)
  {
    p1=qromi(ep,aa,mun,ch,prec); p2=qromo(ep,mun,bb,ch,prec); tetpil=avma;
    return gerepile(av,tetpil,gmulsg(l,gadd(p1,p2)));
  }
  return qromi(ep,a,b,ch,prec);
}

/********************************************************************/
/**                                                                **/
/**                  ZAGIER POLYNOMIALS (for sumiter)              **/
/**                                                                **/
/********************************************************************/

GEN
polzag(long n, long m)
{
  long d1,d,r,k,av=avma,tetpil;
  GEN p1,p2,pol1,g,s;

  if (m>=n || m<0)
    err(talker,"first index must be greater than second in polzag");
  d1=n-m; d=d1<<1; d1--; p1=gsub(gun,gmul2n(polx[0],1));
  pol1=gsub(gun,polx[0]); p2=gmul(polx[0],pol1); r=(m+1)>>1;
  g=gzero;
  for (k=0; k<=d1; k++)
  {
    s=binome(stoi(d),(k<<1)+1); if (k&1) s=negi(s);
    g=gadd(g,gmul(s,gmul(gpuigs(polx[0],k),gpuigs(pol1,d1-k))));
  }
  g=gmul(g,gpuigs(p2,r));
  if (!(m&1)) g=gadd(gmul(p1,g),gmul2n(gmul(p2,deriv(g,0)),1));
  for (k=1; k<=r; k++)
  {
    g=deriv(g,0); g=gadd(gmul(p1,g),gmul2n(gmul(p2,deriv(g,0)),1));
  }
  g = m ? gmul2n(g,(m-1)>>1):gmul2n(g,-1);
  s=mulsi(n-m,mpfact(m+1));
  tetpil=avma; return gerepile(av,tetpil,gdiv(g,s));
}

GEN
polzagreel(long n, long m, long prec)
{
  long d1,d,r,j,k,k2,av=avma,tetpil;
  GEN p2,pol1,g,h,v,b,gend,s,unreel;

  if (m>=n || m<0)
    err(talker,"first index must be greater than second in polzag");
  affsr(1,unreel=cgetr(prec));
  d1=n-m; d=d1<<1; d1--; pol1=gadd(gun,polx[0]);
  p2=gmul(polx[0],pol1); r=(m+1)>>1; gend=stoi(d);
  v=cgetg(d1+2,t_VEC); g=cgetg(d1+2,t_VEC);
  v[d1+1]=un; b=mulri(unreel,gend); g[d1+1]=(long)b;
  for (k=1; k<=d1; k++)
  {
    v[d1-k+1]=un;
    for (j=1; j<k; j++)
      v[d1-k+j+1]=(long)addii((GEN)v[d1-k+j+1],(GEN)v[d1-k+j+2]);
    k2=k+k; b=divri(mulri(b,mulss(d-k2+1,d-k2)),mulss(k2,k2+1));
    for (j=1; j<=k; j++)
      g[d1-k+j+1]=(long)mpadd((GEN)g[d1-k+j+1],mulri(b,(GEN)v[d1-k+j+1]));
    g[d1-k+1]=(long)b;
  }
  h=cgetg(d1+3,t_POL); h[1]=evalsigne(1)+evallgef(d1+3);
  for (k=0; k<=d1; k++) h[k+2]=g[k+1];
  g=gmul(h,gpuigs(p2,r));
  for (j=0; j<=r; j++)
  {
    if (j) g=deriv(g,0);
    if (j || !(m&1))
    {
      h=cgetg(n+3,t_POL); h[1]=evalsigne(1)+evallgef(n+3);
      h[2]=g[2];
      for (k=1; k<n; k++)
	h[k+2]=ladd(gmulsg(k+k+1,(GEN)g[k+2]),gmulsg(k<<1,(GEN)g[k+1]));
      h[n+2]=lmulsg(n<<1,(GEN)g[n+1]); g=h;
    }
  }
  g = m ?gmul2n(g,(m-1)>>1):gmul2n(g,-1);
  s=mulsi(n-m,mpfact(m+1));
  tetpil=avma; return gerepile(av,tetpil,gdiv(g,s));
}

/********************************************************************/
/**                                                                **/
/**            SEARCH FOR REAL ZEROS of an expression              **/
/**                                                                **/
/********************************************************************/

GEN
zbrent(entree *ep, GEN a, GEN b, char *ch, long prec)
{
  long av=avma,tetpil,sig,iter,itmax;
  GEN q1,c,d,e,tol,toli,min1,min2,fa,fb,fc,p,q,r,s,xm;

  if (typ(a)!=t_REAL) { p=cgetr(prec); gaffect(a,p); a=p; }
  if (typ(b)!=t_REAL) { p=cgetr(prec); gaffect(b,p); b=p; }
  sig=cmprr(b,a); if (!sig) { avma = av; return gzero; }
  if (sig<0) { c=a; a=b; b=c; }
  newvalue(ep,a); fa=lisexpr(ch);
  changevalue(ep,b); q1=(GEN)ep->value; fb=lisexpr(ch);
  if (gsigne(fa)*gsigne(fb) > 0)
    err(talker,"roots must be bracketed in solve");
  itmax=(prec<<TWOPOTBITS_IN_LONG)+50; affsr(1,tol=cgetr(3));
  tol=shiftr(tol,-(((prec-2)<<TWOPOTBITS_IN_LONG)-5));
  fc=fb;
  for (iter=1; iter<=itmax; iter++)
  {
    if (gsigne(fb)*gsigne(fc) > 0)
    {
      c=a; fc=fa; d=subrr(b,a); e=d;
    }
    if (gcmp(gabs(fc,0),gabs(fb,0)) < 0)
    {
      a=b; b=c; c=a; fa=fb; fb=fc; fc=fa;
    }
    toli=mulrr(tol,absr(b));
    xm=shiftr(subrr(c,b),-1);
    if ((cmprr(absr(xm),toli)<=0)||gcmp0(fb))
    {
      tetpil=avma;
      killvalue(ep);
      return gerepile(av,tetpil,gcopy(b));
    }
    if (cmprr(absr(e),toli) >= 0 && gcmp(gabs(fa,0),gabs(fb,0)) > 0)
    {
      s=gdiv(fb,fa);
      if (cmprr(a,b)==0)
      {
	p=gmul2n(gmul(xm,s),1); q=gsubsg(1,s);
      }
      else
      {
	q=gdiv(fa,fc); r=gdiv(fb,fc);
	p=gmul2n(gmul(gsub(q,r),gmul(xm,q)),1);
	p=gmul(s,gsub(p,gmul(gsub(b,a),gsubgs(r,1))));
	q=gmul(gmul(gsubgs(q,1),gsubgs(r,1)),gsubgs(s,1));
      }
      if (gsigne(p)>0) q=gneg(q);
      p=gabs(p,0);
      min1=gsub(gmulsg(3,gmul(xm,q)),gabs(gmul(q,toli),0));
      min2=gabs(gmul(e,q),0);
      if (gcmp(gmul2n(p,1),gmin(min1,min2)) < 0) { e=d; d=gdiv(p,q); }
      else { d=xm; e=d; }
    }
    else { d=xm; e=d; }
    a=b; fa=fb;
    if (gcmp(gabs(d,0),toli)>0) b=addrr(b,d);
    else
    {
      if (gsigne(xm)>0) b=addrr(b,absr(toli));
      else b=subrr(b,absr(toli));
    }
    gaffect(b,q1); ; fb=lisexpr(ch);
  }
  err(talker,"too many iterations in solve");
  return NULL; /* not reached */
}

/***********************************************************************/
/*                                                                     */
/*                          SET OPERATIONS                             */
/*                                                                     */
/***********************************************************************/

GEN
gtoset(GEN x)
{
  long tx=typ(x),av=avma,tetpil,c,i,ly;
  GEN y,z;

  if (! is_vec_t(tx)) y=gtovec(x); else y=x;
  ly=lg(y); if (ly==1) { avma=av; return cgetg(1,t_VEC); }
  y=sort(y); z=cgetg(ly,t_VEC);
  c=1; z[1]=y[1];
  for (i=2; i<ly; i++)
    if (!gegal((GEN)y[i],(GEN)z[c])) z[++c]=y[i];
  tetpil=avma; y=cgetg(c+1,t_VEC);
  for (i=1; i<=c; i++) y[i]=lcopy((GEN)z[i]);
  return gerepile(av,tetpil,y);
}

long
isvecset(GEN x)
{
  long lx,i;

  if (typ(x)!=t_VEC) return 0;
  lx=lg(x);
  for (i=1; i<lx-1; i++)
    if (gcmp((GEN)x[i+1],(GEN)x[i])<=0) return 0;
  return 1;
}

/* looks if y belongs to the set x and returns the index if yes, 0 if no */
long setsearch(GEN x, GEN y)
{
  long lx,j,li,ri,fl;

  if (!isvecset(x)) err(talker,"not a set in setsearch");
  lx=lg(x); li=1; ri=lx-1;
  while (ri>=li)
  {
    j = (ri+li)>>1;
    fl=gcmp((GEN)x[j],y);
    if (!fl) return j;
    if (fl<0) li=j+1; else ri=j-1;
  }
  return 0;
}

GEN
setunion(GEN x, GEN y)
{
  long av=avma,tetpil;
  GEN z;

  if (!isvecset(x) || !isvecset(y)) err(talker,"not a set in setunion");
  z=concat(x,y); tetpil=avma; return gerepile(av,tetpil,gtoset(z));
}

GEN
setintersect(GEN x, GEN y)
{
  long av=avma,tetpil,i,lx,c;
  GEN z,p1;

  if (!isvecset(x) || !isvecset(y)) err(talker,"not a set in setintersect");
  lx=lg(x); z=cgetg(lx,t_VEC); c=0;
  for (i=1; i<lx; i++) { p1=(GEN)x[i]; if (setsearch(y,p1)) z[++c]=(long)p1; }
  tetpil=avma; p1=cgetg(c+1,t_VEC);
  for (i=1; i<=c; i++) p1[i]=lcopy((GEN)z[i]);
  return gerepile(av,tetpil,p1);
}

GEN
setminus(GEN x, GEN y)
{
  long av=avma,tetpil,i,lx,c;
  GEN z,p1;

  if (!isvecset(x) || !isvecset(y)) err(talker,"not a set in setintersect");
  lx=lg(x); z=cgetg(lx,t_VEC); c=0;
  for (i=1; i<lx; i++) { p1=(GEN)x[i]; if (!setsearch(y,p1)) z[++c]=(long)p1; }
  tetpil=avma; p1=cgetg(c+1,t_VEC);
  for (i=1; i<=c; i++) p1[i]=lcopy((GEN)z[i]);
  return gerepile(av,tetpil,p1);
}

/***********************************************************************/
/*                                                                     */
/*               OPERATIONS ON DIRICHLET SERIES                        */
/*                                                                     */
/***********************************************************************/

/* Addition, subtraction and scalar multiplication of Dirichlet series
   are done on the corresponding vectors */

static long
dirval(GEN x)
{
  long i=1,lx=lg(x);
  while (i<lx && gcmp0((GEN)x[i])) i++;
  return i;
}

GEN
dirmul(GEN x, GEN y)
{
  long lx,ly,lz,dx,dy,av,tetpil,i,j;
  GEN z,p1;

  if (typ(x)!=t_VEC || typ(y)!=t_VEC) err(talker,"not a dirseries in dirmul");
  av=avma; dx=dirval(x); dy=dirval(y); lx=lg(x); ly=lg(y);
  if (ly-dy<lx-dx) { z=y; y=x; x=z; lz=ly; ly=lx; lx=lz; lz=dy; dy=dx; dx=lz; }
  lz=min(lx*dy,ly*dx);
  z=cgetg(lz,t_VEC); for (i=1; i<lz; i++) z[i]=zero;
  for (j=dx; j<lx; j++)
  {
    p1=(GEN)x[j];
    if (!gcmp0(p1))
    {
      if (gcmp1(p1))
	for (i=j*dy; i<lz; i+=j) z[i]=ladd((GEN)z[i],(GEN)y[i/j]);
      else
      {
	if (gcmp_1(p1))
	  for (i=j*dy; i<lz; i+=j) z[i]=lsub((GEN)z[i],(GEN)y[i/j]);
	else
	  for (i=j*dy; i<lz; i+=j) z[i]=ladd((GEN)z[i],gmul(p1,(GEN)y[i/j]));
      }
    }
  }
  tetpil=avma; return gerepile(av,tetpil,gcopy(z));
}

GEN
dirdiv(GEN x, GEN y)
{
  long lx,ly,lz,dx,dy,av,tetpil,i,j;
  GEN z,p1;

  if (typ(x)!=t_VEC || typ(y)!=t_VEC) err(talker,"not a dirseries in dirmul");
  av=avma; dx=dirval(x); dy=dirval(y); lx=lg(x); ly=lg(y);
  if (dy!=1) err(talker,"not an invertible dirseries in dirdiv");
  lz=min(lx,ly*dx); p1=(GEN)y[1];
  if (!gcmp1(p1)) { y=gdiv(y,p1); x=gdiv(x,p1); }
  else x=gcopy(x);
  z=cgetg(lz,t_VEC); for (i=1; i<dx; i++) z[i]=zero;
  for (j=dx; j<lz; j++)
  {
    p1=(GEN)x[j]; z[j]=(long)p1;
    if (!gcmp0(p1))
    {
      if (gcmp1(p1))
	for (i=j+j; i<lz; i+=j) x[i]=lsub((GEN)x[i],(GEN)y[i/j]);
      else
      {
	if (gcmp_1(p1))
	  for (i=j+j; i<lz; i+=j) x[i]=ladd((GEN)x[i],(GEN)y[i/j]);
	else
	  for (i=j+j; i<lz; i+=j) x[i]=lsub((GEN)x[i],gmul(p1,(GEN)y[i/j]));
      }
    }
  }
  tetpil=avma; return gerepile(av,tetpil,gcopy(z));
}

GEN
dirzetak(GEN nf, GEN b)
{
  long av=avma,tetpil,av1,lim,i,j,lx,bs,c,q,p2s;
  byteptr d=diffptr;
  GEN p,p1,p2,z,z2,pol;

  nf=checknf(nf); pol=(GEN)nf[1]; lim=(avma+bot)>>1;
  if (typ(b)!=t_INT) err(talker,"not an integer type in dirzeta");
  if (signe(b)<=0) bs=2;
  else if (gcmpgs(b,LGBITS)>=0) bs=LGBITS;
  else bs = itos(b)+1;
  p=cgeti(3); p[1]=evalsigne(1)+evallgefint(3); p[2]=2; 
  c = *d++; if (!c) err(recprimer);
  av1=avma;
  z=cgetg(bs,t_VEC); z[1]=un; for (i=2; i<bs; i++) z[i]=zero;
  do
  {
    if (smodis((GEN)nf[4],p[2])) /* p unramified */
      { p1=(GEN)simplefactmod(pol,p)[1]; lx=lg(p1); }
    else
    {
      p2=primedec(nf,p); lx=lg(p2); p1=cgetg(lx,t_COL);
      for (i=1; i<lx; i++) p1[i]=mael(p2,i,4);
    }
    for (j=1; j<lx; j++)
    {
      p2=gpuigs(p,itos((GEN)p1[j])); /* p2 = p^f */
      if (gcmpgs(p2,bs) < 0)
      {
	z2=dummycopy(z); q = p2s = itos(p2);
	while (q<bs)
	{
	  for (i=q; i<bs; i+=q) z2[i] = ladd((GEN)z2[i],(GEN)z[i/q]);
	  q *= p2s;
	}
	z=z2;
      }
    }
    if (avma<lim)
    {
      if(DEBUGMEM>1) fprintferr("Collecting garbage in `dirzetak'\n");
      tetpil=avma; z=gerepile(av1,tetpil,gcopy(z));
    }
    c = *d++; p[2]+=c;
  }
  while (c && p[2]<bs);
  tetpil=avma; return gerepile(av,tetpil,gcopy(z));
}

/*************************************************************************/
/**									**/
/**			         RANDOM					**/
/**									**/
/*************************************************************************/

/* BSD rand gives this: pari_randseed = 1103515245*pari_randseed+12345; */
long
mymyrand()
{
#ifdef LONG_IS_64BIT
  if (BITS_IN_RANDOM==64)
    pari_randseed = 1000000000000654397*pari_randseed + 12347;
  else
    pari_randseed = (1000276549*pari_randseed + 12347) % 2147483648;
#else
  pari_randseed=1000276549*pari_randseed+12347;
  if (pari_randseed<0) pari_randseed += HIGHBIT;
#endif
  return pari_randseed;
}

GEN
genrand(void) { return stoi(mymyrand()); }

GEN
setrand(long seed) { return stoi(pari_randseed = seed); }

GEN
getrand() { return stoi(pari_randseed); }

GEN
getstack() { return stoi(top-avma); }

GEN
gettime() { return stoi(timer2()); }

/***********************************************************************/
/**							              **/
/**       		     PERMUTATIONS                             **/
/**								      **/
/***********************************************************************/

GEN
permute(long n, GEN x)
{
  long av=avma,tetpil,i,a,r;
  GEN v,w,y;

  v=cgetg(n+1,t_VEC); v[1]=1;
  for (r=2; r<=n; r++)
  {
    x=dvmdis(x,r,&w); a=itos(w);
    for (i=r; i>=a+2; i--) v[i]=v[i-1];
    v[a+1]=r;
  }
  tetpil=avma; y=cgetg(n+1,t_VEC);
  for (i=1; i<=n; i++) y[i]=lstoi(v[i]);
  return gerepile(av,tetpil,y);
}

GEN
permuteInv(GEN x)
{
  long av=avma,tetpil, len=lg(x)-1, ini=len, last, ind;
  GEN ary,res;

  if (typ(x)!=t_VEC && typ(x)!=t_COL) err(talker,"not a vector in permuteInv");
  res=gzero; ary=cgetg(len+1,t_VEC);
  for (ind=1; ind<=len; ind++) ary[ind]=*++x;
  ary++;
  for (last=len; last>0; last--)
  {
    len--; ind=len;
    while (ind>0 && itos((GEN)ary[ind])!=last) ind--;
    res=mulis(res,last); tetpil=avma; res=addis(res,ind);
    while (ind++<len) ary[ind-1]=ary[ind];
  }
  if (!signe(res)) { tetpil=avma; res=mpfact(ini); }
  return gerepile(av,tetpil,res);
}

/********************************************************************/
/**                                                                **/
/**                       MODREVERSE                               **/
/**                                                                **/
/********************************************************************/

GEN
polymodrecip(GEN x)
{
  long v,i,j,n,av,tetpil,lx;
  GEN p1,p2,p3,p,phi,y,col;

  if (typ(x)!=t_POLMOD) err(polymoder1);
  p=(GEN)x[1]; phi=(GEN)x[2];
  v=varn(p); n=lgef(p)-3; if (n<=0) return gcopy(x);
  if (n==1)
  {
    y=cgetg(3,t_POLMOD); p1=cgetg(4,t_POL); y[1]=(long)p1;
    p1[1]=p[1]; p1[2]= (typ(phi)==t_POL) ? lneg((GEN)phi[2]) : lneg(phi);
    p1[3]=un; p1=cgetg(3,t_POL); y[2]=(long)p1;
    if (gcmp0((GEN)p[2])) p1[1]=2;
    else
    {
      p1[1]=p[1]-1; av=avma; p2=gdiv((GEN)p[2],(GEN)p[3]);
      tetpil=avma; p1[2]=lpile(av,tetpil,gneg(p2));
    }
    setvarn(p1,v); return y;
  }
  if (gcmp0(phi)) err(polymoder2);
  av=avma; y=cgetg(n+1,t_MAT); p1=cgetg(n+1,t_COL);
  y[1]=(long)p1; p1[1]=un; for (i=2; i<=n; i++) p1[i]=zero;
  p2=phi;
  for (j=2; j<=n; j++)
  {
    lx=lgef(p2); p1=cgetg(n+1,t_COL); y[j]=(long)p1;
    for (i=1; i<=lx-2; i++) p1[i]=p2[i+1];
    for (   ; i<=n; i++) p1[i]=zero;
    if (j<n) p2=gmod(gmul(p2,phi),p);
  }
  col=cgetg(n+1,t_COL); col[1]=zero; col[2]=un;
  for (i=3; i<=n; i++) col[i]=zero;
  p1=gauss(y,col); p2=gtopolyrev(p1,v); p3=caract(x,v);
  tetpil=avma; return gerepile(av,tetpil,gmodulcp(p2,p3));
}
