/*******************************************************************/
/*******************************************************************/
/*                                                                 */
/*                 INTERFACE TO READLINE COMPLETION                */
/*                                                                 */
/*******************************************************************/
/*******************************************************************/
/* $Id: gp_rl.c,v 1.922 1997/04/06 02:52:40 karim Exp karim $ */
#include "genpari.h"
#include "anal.h"
#include "gp.h"

#ifdef READLINE
BEGINEXTERN
#include <readline.h>
char **pari_completion(char *text, int start, int end);
char *filename_completion_function(char *text,int state);
char *username_completion_function(char *text,int state);

int rl_completion_query_items;
int rl_bind_key_in_map (int key, Function *function, Keymap keymap);
ENDEXTERN

void print_fun_list(char **matches, int nbli);
void external_help(char *s);

extern default_type gp_default_list[], gp_member_list[];
static int under_emacs;
static entree *current_ep = NULL;

typedef char** (*CF)(char*, char * (*)()); /* completion function */
typedef char* (*GF)(char*,int); /* generator function */

/*
 * Attempt to complete on the contents of TEXT. START and END show the
 * region of TEXT that contains the word to complete.  We can use the
 * entire line in case we want to do some simple parsing.  Return the
 * array of matches, or NULL if there aren't any.
 */
char **
get_matches(int end, char *text, char* f(char*,int))
{
  char **matches = ((CF) completion_matches)(text, (char *(*)(ANYARG))f);

  if (matches && ! matches[1] /* only one match          */
	      && ! current_ep /* this is not a variable  */
              && end >= 0     /* from command_completion */
              && rl_line_buffer[end] != '(' )
  {
    int i=strlen(matches[0]);
    matches[0] = (char*) xrealloc(matches[0], i+2, i);
    strcat(matches[0],"(");
  }
  current_ep = NULL;

  if (under_emacs)
  {
    int i;
    if (!matches) printf("@");
    else
    {
      printf("%s@", matches[0] + strlen(text));
      if (matches[1]) print_fun_list(matches+1,0);

     /* we don't want readline to do anything, but insert some junk
      * which will be erased by emacs.
      */
      for (i=0; matches[i]; i++) free(matches[i]);
      free(matches);
    }
    matches = (char **) xmalloc(2*sizeof(char *));
    matches[0] = xmalloc(2); sprintf(matches[0],"_");
    matches[1] = NULL;
    printf("@E_N_D"); fflush(stdout);
  }
  return matches;
}

/*
 * Generator function for command completion.  STATE lets us know whether
 * to start from scratch; without any state (i.e. STATE == 0), then we
 * start at the top of the list.
 */
char *
command_generator (char *text, int  state)
{
  static int hashpos, len, junk;
  static entree* ep;
  char *name;

 /*
  * If this is a new word to complete, initialize now.  This includes
  * saving the length of TEXT for efficiency, and initializing the index
  * variable to 0. Since file completion and symbol completion use
  * different word boundaries, put a new boundary into junk.
  */
  if (!state)
  {
    hashpos=0; ep=functions_hash[hashpos]; len=strlen (text); junk=len-1;
    while (junk >= 0 && isalnum(text[junk])) junk--;  
    junk++; len -= junk;
  }

  /* Return the next name which partially matches from the command list. */
  while (!(ep && !strncmp(ep->name,text+junk,len))
       && (ep||++hashpos<functions_tblsz))
    ep = ep ? ep->next: functions_hash[hashpos];
  if (ep)
  {
    if (EpVALENCE(ep)==EpVAR) current_ep = ep;
    name = strncpy((char*) xmalloc(strlen(ep->name) + 1 + junk),text,junk);
    strcpy(name+junk,ep->name); ep=ep->next;
    return(name);
  }
  return NULL; /* no names matched */
}

char *
member_generator (char *text, int  state)
{
  static int n, len, junk;
  char *name, *def = NULL;

  if (!state)
  {
    n = 0; len = strlen (text); junk=len-1;
    while (junk >= 0 && isalnum(text[junk])) junk--;  
    junk++; len -= junk;
  }

  do
    def = gp_member_list[n++].name;
  while ( def && strncmp(def,text+junk,len) );

  if (def)
  {
    name = strncpy((char*) xmalloc(strlen(def) + 1 + junk),text,junk);
    strcpy(name+junk,def); return(name);
  }
  return NULL; /* no names matched */
}

#define DFLT 0
#define ENTREE 1

char *
generator(void *list, char *text, int *nn, int len, int typ)
{
  char *def = NULL, *name;
  int n = *nn;

  /* Return the next name which partially matches from list.*/
  switch(typ)
  {
    case DFLT : 
      do
	def = (((default_type *) list)[n++]).name;
      while ( def && strncmp(def,text,len) );
      break;

    case ENTREE :
      do
	def = (((entree *) list)[n++]).name;
      while ( def && strncmp(def,text,len) );
  }

  *nn = n;
  if (def)
  {
    name = strcpy((char*) xmalloc(strlen(def)+1), def);
    return name;
  }
  return NULL; /* no names matched */
}

char *
old_generator(char *text,int state)
{
  static int n,len;
  static char *res;

  if (!state) { res = "a"; n=0; len=strlen(text); }
  if (res)
  {
    res = generator((void *)oldfonctions,text,&n,len,ENTREE);
    if (res) return res;
    n=0;
  }
  return generator((void *)functions_oldgp,text,&n,len,ENTREE);
}

char *
default_generator (char *text,int state)
{
  static int n,len;

  if (!state) { n=0; len=strlen(text); }
  return generator(gp_default_list,text,&n,len,DFLT);
}

char **
pari_completion (char *text, int start, int end)
{
  char *s, *beg;
  int i, first=0;
/*
 * If the line does not begin by a backslash, then it is:
 * . an old command ( if preceded by "whatnow(" ).
 * . a default ( if preceded by "default(" ).
 * . a member function ( if preceded by "." within 4 letters )
 * . a file name (in current directory) ( if preceded by "read(" )
 * . a command 
 *
 *   Last by default
 */
  if (start >=1 && rl_line_buffer[start] != '~') start--;
  while (start && isalnum(rl_line_buffer[start])) start--;
  if (rl_line_buffer[start] == '~')
  {
    for(i=start+1;i<=end;i++)
      if (rl_line_buffer[i] == '/')
	return get_matches(-1,text,filename_completion_function);
    return get_matches(-1,text,username_completion_function);
  }

  while (rl_line_buffer[first] && isspace(rl_line_buffer[first])) first++;
  switch (rl_line_buffer[first])
  {
    case '\\':
      if (first == start) text++;
      return get_matches(-1,text,filename_completion_function);
    case '?':
      return get_matches(-1,text,command_generator);
  }

  while (start && rl_line_buffer[start] != '('
               && rl_line_buffer[start] != ',') start--;
  if (rl_line_buffer[start] == '(')
  {
    beg = rl_line_buffer + start;
    if (start >= 7)
    {
      s = beg-7;
      if (!strncmp(s,"default",7))
	return get_matches(-1,text,default_generator);
      if (!strncmp(s,"whatnow",7))
	return get_matches(-1,text,old_generator);
    }

    if (start>=4)
    {
      s = beg-4;
      if (!strncmp(s,"read",4))
	return get_matches(-1,text,filename_completion_function);
    }
  }
  for(i=end-1;i>=start;i--)
    if (rl_line_buffer[i] == '.')
      return get_matches (-1,text,member_generator);

  return get_matches(end,text,command_generator);
}

int
rl_short_help(int count, int key)
{
  long hashpos;
  int off=rl_point, point=off, len;
  entree *ep;
  char *u, *old;
  FILE *save = outfile;

  while (off && isalnum(rl_line_buffer[off-1])) off--;
  old = u = rl_line_buffer+off; hashpos = hashvalue(&u,functions_tblsz,1);
  len = u - old;
  for (ep = functions_hash[hashpos]; ep; ep = ep->next)
    if (len == strlen(ep->name) && !strncmp(ep->name, old, len))
      break;
  if (!ep || !ep->help) { ding(); return 0;}

  rl_point = 0; outfile = rl_outstream;
  pariputc('\n');

  if (count>0)
    print_help(ep->help);
  else
    external_help(ep->name);
  rl_point = point;
  rl_refresh_line();

  outfile = save; return 0;
}

int
rl_long_help(int count, int key)
{
  return rl_short_help(-1,key);
}

void
init_readline(int i)
{
  /* Allow conditional parsing of the ~/.inputrc file. */
  rl_readline_name = "Pari-GP";

  /* added ? and , */
  rl_basic_word_break_characters=" \t\n\"\\'`@$><=;|&{(?,";

  rl_special_prefixes="~";

  /* custom completer */
#ifdef CPPFunction_defined
  rl_attempted_completion_function = (CPPFunction *) pari_completion;
#else
  rl_attempted_completion_function = (Function *) pari_completion;
#endif

  /* we always want the whole list of completions under emacs */
  under_emacs=i; if (under_emacs) rl_completion_query_items = 0x8fff;

#  define Bind ((void(*)(int,Function*,Keymap)) rl_bind_key_in_map)
#  define Defun ((void(*)(const char*,Function*,int)) rl_add_defun)

  Defun("short-help", (Function*) rl_short_help, -1);
  Defun("long-help", (Function*) rl_long_help, -1);

  Bind('h', (Function*) rl_short_help, emacs_meta_keymap);
  Bind('H', (Function*) rl_long_help, emacs_meta_keymap);
  Bind('h', (Function*) rl_short_help, vi_movement_keymap);
  Bind('H', (Function*) rl_long_help, vi_movement_keymap);

#  ifdef EMACS_DOS_KEYMAP
     Bind(';', (char*) rl_short_help, emacs_dos_keymap); /* F1 */
     Bind('T', (char*) rl_long_help, emacs_dos_keymap); /* Shift-F1 */
#  endif
}
#endif
