/*******************************************************************/
/*******************************************************************/
/*                                                                 */
/*            Analyseur syntactique pour la calculette             */
/*                                                                 */
/*                       copyright Babe Cool                       */
/*                                                                 */
/*******************************************************************/
/*******************************************************************/
/* $Id: anal.c,v 1.922 1997/04/06 02:52:40 karim Exp karim $ */
#include "genpari.h"
#include "anal.h"

#define separe(c) ((c)==';' || (c)==':')
typedef GEN (*PFGEN)(ANYARG);

GEN    geni(void);

static GEN   extract0(GEN x, GEN l1, GEN l2);
static GEN   gshiftminus(GEN x, GEN n);
static GEN   gshiftplus(GEN x, GEN n);
static void  alias0(char *s, char *old);
static void  break0(long n);
static void  err_new_fun();
static void  next0();
static void  return0(GEN x);

static GEN    constante(void);
static GEN    expr(void);
static GEN    facteur(void);
static GEN    identifier(void);
static GEN    matrix_block(GEN p, entree *ep);
static GEN    read_member(GEN x);
static GEN    seq(void);
static GEN    truc(void);
static long   number(long *nb);
static char*  doskipseq(char *s);
static void   skipconstante(void);
static void   skipexpr(void);
static void   skipfacteur(void);
static void   skipidentifier(void);
static void   skipseq(void);
static void   skipstring(void);
static void   skiptruc(void);
static entree *entry(void);
static entree *installep(void *f, char *name, int l, int v, int add, int n, entree **table);
static entree *skipentry(void);

typedef struct gpmark
{
  char *identifier;
  char *symbol;
  char *raw;
  char *member;
  char *start;
} gpmark; /* last time we began parsing an object of specified type */

static char *analyseur, *labellist[100];
static long analyseurtetpil, skipping_fun_def;
static entree *check_new_fun;
static gpmark mark;

/*
 * Def: ' ', '\t' and '\n' and "\\" are forbidden
 *      internally (suppressed by filtre).
 *      { } are forbidden everywhere and will be used to denote optional
 *      lexemes in the sequel. The sequence { a }* means any number
 *      (possibly 0) of object a. { x|y } means an optional x or y.
 *      rem: Only "seq" can be empty.
 *
 *   seq :
 *      sequence of { expr{ :|; } }*
 *
 *   expr :
 *      expression = sequence of "facteurs" separated by binary operators
 *      whose priority are:
 *       1: *, /, \, \/, %, >>, <<                (highest)
 *       2: +, -
 *       3: <, <=, >, >=, !=, ==, <>
 *       4: &, &&, |, ||                  (lowest)
 *      read from left to right.
 *
 *   facteur :
 *       Optionnal leading sign (meaningfull only when "facteur" is
 *       enclosed in parentheses), followed by a "truc", then by
 *       any succession of the following:
 *         ~, _, ', !
 *    or   ^ facteur
 *    or   matrix_block
 *    or   .member      (see gp_member_list)
 *
 *   truc:
 *       identifier
 *  or   constante
 *  or   ! truc
 *  or   matrix_block (no_affect=1)
 *  or   (expr)
 *  or   %{ ` }*  or %number
 *
 *   identifier:
 *       entry ( { expr } { ,expr }* )
 *       The () are optional when arg list is void.
 *
 *   matrix_block :
 *       [ A { { ; }A }*] where A = { expr } { { , }{ expr } }*
 *       All A must share the same length.
 *       If (no_affect=0 || ep !=NULL): follows an optionnal "= expr"
 *    or ++, --, op= where op is one of the operators in expr 1: and 2:
 *
 *   entry :
 *       any succesion of alphanumeric characters, the first of which is not
 *       a digit.
 *
 *   constante:
 *       number { . } { number } { e|E } { +|- } { number }.
 *
 *   number:
 *       any non-negative integer.
 */

/* Do not affect (analyseur,analyseurtetpil,mark.start) */
static GEN
lisseq0(char *t, int do_seq)
{
  const long av = avma, oldtetpil = analyseurtetpil;
  char *olds = analyseur, *olde = mark.start;
  GEN x,res;

  if (foreignExprHandler && *t == foreignExprSwitch)
      return (*foreignExprHandler)(t);

  check_new_fun=NULL; skipping_fun_def=0;
  mark.start = analyseur = t; analyseurtetpil = avma;
  res = do_seq? seq(): expr();
  if (check_break_status(NOACTION,&x))
  {
    analyseurtetpil=avma; res = forcecopy(x);
  }
  res = gerepile(av,analyseurtetpil,res);
  analyseur = olds; mark.start = olde; analyseurtetpil = oldtetpil;
  return res;
}

GEN
lisseq(char *t)
{
  return lisseq0(t,1);
}

GEN
lisexpr(char *t)
{
  return lisseq0(t,0);
}

/* filtered lisexpr = remove blanks and comments */
GEN
flisexpr(char *t)
{
  char *tmp = xmalloc(1 + strlen(t));
  GEN x;

  strcpy(tmp,t); filtre(tmp);
  x = lisseq0(tmp,0); free(tmp);
  return x;
}

/* [Ptitboul] NB: n'est utilis que dans gp.c ! */
GEN
readseq(char *c)
{
  GEN res;
  char *nextc;
  long i;

  for (i=0; i<100; i++) labellist[i]=NULL;
  check_new_fun=NULL; skipping_fun_def=0;

  nextc = doskipseq(c); res = lisseq(c);
  if (*nextc) err(talker2,"unused characters: ", nextc, c);
  return res;
}

/* return hashing value for identifier pointed by *u
 * if do_update is non-null, update *u
 * tblsz contains the size of the hashtable
 */ 
long
hashvalue(char **u, int tblsz, int do_update)
{
  char *s=*u;
  long n;

  for (n=0; isalnum(*s); s++) n = (n<<1) ^ *s;
  if (do_update) *u=s;
  if (n < 0) n = -n;
  return n % tblsz;
}

entree *
install(void *f, char *name, int valence)
{
  const int hashpos = hashvalue(&analyseur,functions_tblsz,0);
  entree *ep;

  for (ep = functions_hash[hashpos]; ep; ep = ep->next)
    if (!strcmp(name, ep->name)) err(nomer1);
  if (valence < 0 || valence > 3) err(valencer1);
  return installep(f, name, strlen(name), valence, 0, hashpos, functions_hash);
}

static void
free_args(GEN *x)
{
  GEN *y=x;

  while (*x) killbloc(*x++);
  free((void *)y);
}

void
freeep(entree *ep)
{
  if (foreignFuncFree && ep->code && (*ep->code == 'x'))
    (*foreignFuncFree)(ep); /* function created by foreign interpreter */

  if (ep-> valence & EpSTATIC) return;
  if (ep->help) free(ep->help);
  if (ep->args) free_args(ep->args);
  free(ep);
}

/*
 * Kill entree ep. This means :
 *   variable : free ep->value. return monomial of degree 1 in ep->name.
 *   function/alias : remove ep from hashing table.
 */
GEN
kill0(entree *ep)
{
  int n;
  entree *ep1;

  if (EpVALENCE(ep) < EpUSER) err(killer1,mark.symbol,mark.start);
  if (EpVALENCE(ep) != EpALIAS) killvalue(ep);
  if (EpVALENCE(ep) == EpVAR) return (GEN) ep->value;

  n = hashvalue(&ep->name,functions_tblsz,0);
  if (functions_hash[n] == ep)
  {
    functions_hash[n] = ep->next;
    freeep(ep); return gnil;
  }

  for (ep1 = functions_hash[n]; ep1; ep1 = ep1->next)
  {
    if (ep1->next == ep)
    {
      ep1->next = ep->next;
      freeep(ep); break;
    }
  }
  return gnil;
}

/*
 * List of GP functions:
 * ---------------------
 * Format (struct entree) :
 * {
 *   char *name    : name (under GP).
 *   ulong valence : used to form arg list, now handled by code save for
 *                   some special cases.
 *   void *value   : C function to call.
 *   long menu     : which help section do we belong (See below).
 *   char *code    : argument list (See below).
 *   entree *next  : next entree (init to NULL, used in hashing code).
 *   char *help    : short help text (init to NULL).
 *   GEN *args     : default arguments (NULL terminated).
 * }
 *
 * menu:
 * -----
 *  1: Standard monadic or dyadic OPERATORS
 *  2: CONVERSIONS and similar elementary functions
 *  3: TRANSCENDENTAL functions
 *  4: NUMBER THEORETICAL functions
 *  5: Functions related to ELLIPTIC CURVES
 *  6: Functions related to general NUMBER FIELDS
 *  7: POLYNOMIALS and power series
 *  8: Vectors, matrices, LINEAR ALGEBRA and sets
 *  9: SUMS, products, integrals and similar functions
 *  10: GRAPHIC functions
 *  11: PROGRAMMING under GP
 *
 *
 * code: NULL = use valence instead.
 * -----
 *  I Input position (should be processed with lisexpr or lisseq only).
 *
 *  G Argument is GEN
 *  L Argument is long
 *  S Argument is a symbol (i.e GP function name)
 *  V Argument is a variable (same as S, but valence must equal EpVAR)
 *  n Argument is var number
 *  r Argument is raw input (treated as a string without quotes).
 *    Quoted args are copied as strings. Stops at first unquoted ')' or ','.
 *    Special chars can be quoted using '\'.  Ex : aa"b\n)"c => "aab\n)c".
 *  s Argument is expanded string. Example: pi"x"2 yields "3.142x2".
 *    The unquoted components can be of any pari type (converted according to
 *    the current output format)
 *  s* Print any number of args whose type was s
 *  s*p idem, setting prettyp=1
 *  s*t idem, in TeX format.
 *
 *  F Fake *GEN argument (function requires a *GEN, but we don't use the
 *    resulting GEN)
 *  f Fake *long argument
 *  p arg is precision (to the C library)
 *  P arg is series precision
 *
 *  l Return long
 *  v Return void
 *
 *  = Separator '=' required.
 *
 *  D Has a default value. Format is "Dvalue,type," (the ending comma is
 *    mandatory). Ex: D0,G, (arg is GEN, gzero by default).
 *    The user-given args are read first, then completed by the defaults
 *    (C++ behaviour).
 *
 *  x Installed foreign function. Put the ep of the function as an argument,
 *  then call installedHandler (???).
 *
 * When new codes are added, changes should be made to identifier and
 * skipidentifier.
 *
 * Currently the following functions have no code word: O 50 goto 61 if 80
 * label 60 until 82 while 81
 *
 * Unsupported types: 50, 60-61, 80-82.
 *
 * Valence 99 reserved for functions without the optimized indentifier code.
 * Valence 0 reserved for functions without mandatory args.
 */

entree functions_basic[]={
{ "Euler",0,(void*)mpeuler,3,"p",NULL,NULL,NULL },
{ "I",0,(void*)geni,2,"",NULL,NULL,NULL },
{ "Mat",0,(void*)gtomat,2,"D[],G,",NULL,NULL,NULL },
{ "Mod",25,(void*)Mod0,2,"GGD0,L,",NULL,NULL,NULL },
{ "O",50,NULL,7,NULL,NULL,NULL,NULL },
{ "Pi",0,(void*)mppi,3,"p",NULL,NULL,NULL },
{ "Pol",14,(void*)gtopoly,2,"GDx,n,",NULL,NULL,NULL },
{ "Polrev",14,(void*)gtopolyrev,2,"GDx,n,",NULL,NULL,NULL },
{ "Qfb",3,(void*)Qfb0,2,"GGGD0,G,p",NULL,NULL,NULL },
{ "Ser",14,(void*)gtoser,2,"GDx,n,",NULL,NULL,NULL },
{ "Set",0,(void*)gtoset,2,"D[],G,",NULL,NULL,NULL },
{ "Vec",0,(void*)gtovec,2,"D[],G,",NULL,NULL,NULL },
{ "abs",1,(void*)gabs,3,"Gp",NULL,NULL,NULL },
{ "acos",1,(void*)gacos,3,"Gp",NULL,NULL,NULL },
{ "acosh",1,(void*)gach,3,"Gp",NULL,NULL,NULL },
{ "addprimes",0,(void*)addprimestotable,4,"D[],G,",NULL,NULL,NULL },
{ "agm",2,(void*)agm,3,"GGp",NULL,NULL,NULL },
{ "algdep",23,(void*)algdep0,8,"GLD0,L,p",NULL,NULL,NULL },
{ "alias",2,(void*)alias0,11,"vrr",NULL,NULL,NULL },
{ "arg",1,(void*)garg,3,"Gp",NULL,NULL,NULL },
{ "asin",1,(void*)gasin,3,"Gp",NULL,NULL,NULL },
{ "asinh",1,(void*)gash,3,"Gp",NULL,NULL,NULL },
{ "atan",1,(void*)gatan,3,"Gp",NULL,NULL,NULL },
{ "atanh",1,(void*)gath,3,"Gp",NULL,NULL,NULL },
{ "bernreal",11,(void*)bernreal,3,"Lp",NULL,NULL,NULL },
{ "bernvec",11,(void*)bernvec,3,"L",NULL,NULL,NULL },
{ "besseljh",2,(void*)jbesselh,3,"GGp",NULL,NULL,NULL },
{ "besselk",2,(void*)kbessel0,3,"GGD0,L,p",NULL,NULL,NULL },
{ "bestappr",2,(void*)bestappr,4,"GG",NULL,NULL,NULL },
{ "bezout",2,(void*)vecbezout,4,"GG",NULL,NULL,NULL },
{ "bezoutres",2,(void*)vecbezoutres,4,"GG",NULL,NULL,NULL },
{ "bigomega",1,(void*)gbigomega,4,"G",NULL,NULL,NULL },
{ "binary",1,(void*)binaire,2,"G",NULL,NULL,NULL },
{ "binomial",21,(void*)binome,4,"GL",NULL,NULL,NULL },
{ "bittest",2,(void*)gbittest,2,"GG",NULL,NULL,NULL },
{ "bnfcertify",10,(void*)certifybuchall,6,"Gl",NULL,NULL,NULL },
{ "bnfclassunit",1,(void*)bnfclassunit0,6,"GD0,L,D0,G,p",NULL,NULL,NULL },
{ "bnfclgp",1,(void*)classgrouponly,6,"GD0,G,p",NULL,NULL,NULL },
{ "bnfconductor",62,(void*)bnfconductor,6,"GD0,G,D0,G,D0,L,p",NULL,NULL,NULL },
{ "bnfconductorofchar",2,(void*)bnfconductorofchar,6,"GGp",NULL,NULL,NULL },
{ "bnfdecodemodule",2,(void*)decodemodule,6,"GG",NULL,NULL,NULL },
{ "bnfinit",91,(void*)bnfinit0,6,"GD0,L,D0,G,p",NULL,NULL,NULL },
{ "bnfisintnorm",3,(void*)bnfisintnorm,6,"GG",NULL,NULL,NULL },
{ "bnfisnorm",3,(void*)bnfisnorm,6,"GGD1,L,p",NULL,NULL,NULL },
{ "bnfisprincipal",2,(void*)isprincipalall,6,"GGD1,L,",NULL,NULL,NULL },
{ "bnfissunit",2,(void*)bnfissunit,6,"GGG",NULL,NULL,NULL },
{ "bnfisunit",2,(void*)isunit,6,"GG",NULL,NULL,NULL },
{ "bnfmake",1,(void*)makebigbnf,6,"Gp",NULL,NULL,NULL },
{ "bnfnarrow",91,(void*)buchnarrow,6,"G",NULL,NULL,NULL },
{ "bnfreg",1,(void*)regulator,6,"GD0,G,p",NULL,NULL,NULL },
{ "bnfsignunit",1,(void*)signunit,6,"G",NULL,NULL,NULL },
{ "bnfsunit",1,(void*)bnfsunit,6,"GGp",NULL,NULL,NULL },
{ "bnfunit",1,(void*)buchfu,6,"Gp",NULL,NULL,NULL },
{ "bnrclass",2,(void*)bnrclass0,6,"GGD0,L,p",NULL,NULL,NULL },
{ "bnrclassno",2,(void*)rayclassno,6,"GG",NULL,NULL,NULL },
{ "bnrclassnolist",2,(void*)rayclassnolist,6,"GG",NULL,NULL,NULL },
{ "bnrdisc",62,(void*)bnrdisc0,6,"GD0,G,D0,G,D0,L,p",NULL,NULL,NULL },
{ "bnrdisclist",32,(void*)bnrdisclist0,6,"GGD0,G,D0,L,",NULL,NULL,NULL },
{ "bnrinit",2,(void*)bnrinit0,6,"GGD0,L,p",NULL,NULL,NULL },
{ "bnrisprincipal",2,(void*)isprincipalrayall,6,"GGD1,L,",NULL,NULL,NULL },
{ "break",0,(void*)break0,11,"vD1,L,",NULL,NULL,NULL },
{ "ceil",1,(void*)gceil,2,"G",NULL,NULL,NULL },
{ "centerlift",1,(void*)centerlift,2,"G",NULL,NULL,NULL },
{ "changevar",2,(void*)changevar,2,"GG",NULL,NULL,NULL },
{ "charpoly",14,(void*)charpoly0,8,"GDx,n,D0,L,",NULL,NULL,NULL },
{ "chinese",2,(void*)chinois,4,"GG",NULL,NULL,NULL },
{ "component",21,(void*)compo,2,"GL",NULL,NULL,NULL },
{ "conj",1,(void*)gconj,2,"G",NULL,NULL,NULL },
{ "conjvec",1,(void*)conjvec,2,"Gp",NULL,NULL,NULL },
{ "content",1,(void*)content,4,"G",NULL,NULL,NULL },
{ "contfrac",1,(void*)sfcont0,4,"GD0,G,D0,L,",NULL,NULL,NULL },
{ "contfracpnqn",1,(void*)pnqn,4,"G",NULL,NULL,NULL },
{ "core",1,(void*)core0,4,"GD0,L,",NULL,NULL,NULL },
{ "coredisc",1,(void*)coredisc0,4,"GD0,L,",NULL,NULL,NULL },
{ "cos",1,(void*)gcos,3,"Gp",NULL,NULL,NULL },
{ "cosh",1,(void*)gch,3,"Gp",NULL,NULL,NULL },
{ "decodefactor",1,(void*)decodefactor,4,"G",NULL,NULL,NULL },
{ "denominator",1,(void*)denom,2,"G",NULL,NULL,NULL },
{ "deriv",14,(void*)deriv,7,"Gn",NULL,NULL,NULL },
{ "dilog",1,(void*)dilog,3,"Gp",NULL,NULL,NULL },
{ "dirdiv",2,(void*)dirdiv,7,"GG",NULL,NULL,NULL },
{ "direuler",83,(void*)direuler,7,"V=GGI",NULL,NULL,NULL },
{ "dirmul",2,(void*)dirmul,7,"GG",NULL,NULL,NULL },
{ "dirzetak",2,(void*)dirzetak,6,"GG",NULL,NULL,NULL },
{ "divisors",1,(void*)divisors,4,"G",NULL,NULL,NULL },
{ "divrem",2,(void*)gdiventres,1,"GG",NULL,NULL,NULL },
{ "eint1",1,(void*)eint1,3,"Gp",NULL,NULL,NULL },
{ "elladd",3,(void*)addell,5,"GGG",NULL,NULL,NULL },
{ "ellak",2,(void*)akell,5,"GG",NULL,NULL,NULL },
{ "ellan",23,(void*)anell,5,"GL",NULL,NULL,NULL },
{ "ellap",2,(void*)ellap0,5,"GGD0,L,",NULL,NULL,NULL },
{ "ellbil",3,(void*)bilhell,5,"GGGp",NULL,NULL,NULL },
{ "ellchangecurve",2,(void*)coordch,5,"GG",NULL,NULL,NULL },
{ "ellchangepoint",2,(void*)pointch,5,"GG",NULL,NULL,NULL },
{ "ellglobalred",1,(void*)globalreduction,5,"G",NULL,NULL,NULL },
{ "ellheight",2,(void*)ellheight0,5,"GGD0,L,p",NULL,NULL,NULL },
{ "ellheightmatrix",2,(void*)mathell,5,"GGp",NULL,NULL,NULL },
{ "ellinit",1,(void*)ellinit0,5,"GD0,L,p",NULL,NULL,NULL },
{ "ellisoncurve",20,(void*)oncurve,5,"GGl",NULL,NULL,NULL },
{ "ellj",1,(void*)jell,5,"Gp",NULL,NULL,NULL },
{ "elllocalred",2,(void*)localreduction,5,"GG",NULL,NULL,NULL },
{ "elllseries",4,(void*)lseriesell,5,"GGGGp",NULL,NULL,NULL },
{ "ellorder",2,(void*)orderell,5,"GG",NULL,NULL,NULL },
{ "ellordinate",2,(void*)ordell,5,"GGp",NULL,NULL,NULL },
{ "ellpointtoz",2,(void*)zell,5,"GGp",NULL,NULL,NULL },
{ "ellpow",3,(void*)powell,5,"GGGp",NULL,NULL,NULL },
{ "ellsub",3,(void*)subell,5,"GGGp",NULL,NULL,NULL },
{ "elltaniyama",1,(void*)taniyama,5,"Gp",NULL,NULL,NULL },
{ "elltors",1,(void*)torsell,5,"Gp",NULL,NULL,NULL },
{ "ellwp",1,(void*)weipell,5,"G",NULL,NULL,NULL },
{ "ellztopoint",2,(void*)pointell,5,"GGp",NULL,NULL,NULL },
{ "erfc",1,(void*)gerfc,3,"Gp",NULL,NULL,NULL },
{ "eta",1,(void*)eta,3,"Gp",NULL,NULL,NULL },
{ "eulerphi",1,(void*)gphi,4,"G",NULL,NULL,NULL },
{ "eval",1,(void*)geval,7,"G",NULL,NULL,NULL },
{ "exp",1,(void*)gexp,3,"Gp",NULL,NULL,NULL },
{ "factor",1,(void*)factor0,4,"GD-1,L,",NULL,NULL,NULL },
{ "factorcantor",2,(void*)factcantor,4,"GG",NULL,NULL,NULL },
{ "factorff",3,(void*)factmod9,4,"GGG",NULL,NULL,NULL },
{ "factorial",11,(void*)mpfactr,4,"Lp",NULL,NULL,NULL },
{ "factormod",2,(void*)factormod0,4,"GGD0,L,",NULL,NULL,NULL },
{ "factornf",2,(void*)polfnf,6,"GG",NULL,NULL,NULL },
{ "factorpadic",32,(void*)factorpadic4,7,"GGLD0,L,",NULL,NULL,NULL },
{ "ffinit",21,(void*)Fpn,4,"GL",NULL,NULL,NULL },
{ "fibonacci",11,(void*)fibo,4,"L",NULL,NULL,NULL },
{ "floor",1,(void*)gfloor,2,"G",NULL,NULL,NULL },
{ "for",83,(void*)forpari,11,"vV=GGI",NULL,NULL,NULL },
{ "fordiv",84,(void*)fordiv0,11,"vGVI",NULL,NULL,NULL },
{ "forprime",83,(void*)forprime,11,"vV=GGI",NULL,NULL,NULL },
{ "forstep",86,(void*)forstep,11,"vV=GGGI",NULL,NULL,NULL },
{ "forvec",87,(void*)forvec,11,"vV=GI",NULL,NULL,NULL },
{ "frac",1,(void*)gfrac,2,"G",NULL,NULL,NULL },
{ "gamma",1,(void*)ggamma,3,"Gp",NULL,NULL,NULL },
{ "gammah",1,(void*)ggamd,3,"Gp",NULL,NULL,NULL },
{ "gcd",2,(void*)gcd0,4,"GGD0,L,",NULL,NULL,NULL },
{ "getheap",0,(void*)getheap,11,"",NULL,NULL,NULL },
{ "getrand",0,(void*)getrand,11,"",NULL,NULL,NULL },
{ "getstack",0,(void*)getstack,11,"",NULL,NULL,NULL },
{ "gettime",0,(void*)gettime,11,"",NULL,NULL,NULL },
{ "goto",61,NULL,11,NULL,NULL,NULL,NULL },
{ "hilbert",30,(void*)hil0,4,"GGD1,G,l",NULL,NULL,NULL },
{ "hyperu",3,(void*)hyperu,3,"GGGp",NULL,NULL,NULL },
{ "idealadd",3,(void*)idealadd,6,"GGG",NULL,NULL,NULL },
{ "idealaddone",3,(void*)idealaddone0,6,"GGD0,G,",NULL,NULL,NULL },
{ "idealappr",2,(void*)idealappr0,6,"GGD0,L,",NULL,NULL,NULL },
{ "idealchinese",3,(void*)idealchinese,6,"GGG",NULL,NULL,NULL },
{ "idealcoprime",3,(void*)idealcoprime,6,"GGG",NULL,NULL,NULL },
{ "idealdiv",3,(void*)idealdiv0,6,"GGGD0,L,",NULL,NULL,NULL },
{ "idealfactor",2,(void*)idealfactor,6,"GG",NULL,NULL,NULL },
{ "idealhnf",2,(void*)idealmathnf0,6,"GGD0,G,",NULL,NULL,NULL },
{ "idealintersect",3,(void*)idealintersect,6,"GGG",NULL,NULL,NULL },
{ "idealinv",2,(void*)idealinv0,6,"GGD0,L,",NULL,NULL,NULL },
{ "ideallist",21,(void*)ideallist0,6,"GLD4,L,",NULL,NULL,NULL },
{ "ideallistarch",3,(void*)ideallistarch0,6,"GGD0,G,D0,L,",NULL,NULL,NULL },
{ "ideallog",3,(void*)zideallog,6,"GGGp",NULL,NULL,NULL },
{ "idealmin",3,(void*)minideal,6,"GGGp",NULL,NULL,NULL },
{ "idealmul",3,(void*)idealmul0,6,"GGGD0,L,p",NULL,NULL,NULL },
{ "idealnorm",2,(void*)idealnorm,6,"GG",NULL,NULL,NULL },
{ "idealpow",3,(void*)idealpow0,6,"GGGD0,L,p",NULL,NULL,NULL },
{ "idealprimedec",2,(void*)primedec,6,"GGp",NULL,NULL,NULL },
{ "idealprincipal",2,(void*)principalideal,6,"GG",NULL,NULL,NULL },
{ "idealred",3,(void*)ideallllred,6,"GGGp",NULL,NULL,NULL },
{ "idealstar",2,(void*)idealstar0,6,"GGD1,L,",NULL,NULL,NULL },
{ "idealtwoelt",2,(void*)ideal_two_elt0,6,"GGD0,G,",NULL,NULL,NULL },
{ "idealval",30,(void*)idealval,6,"GGGl",NULL,NULL,NULL },
{ "ideleprincipal",2,(void*)principalidele,6,"GG",NULL,NULL,NULL },
{ "if",80,NULL,11,NULL,NULL,NULL,NULL },
{ "imag",1,(void*)gimag,2,"G",NULL,NULL,NULL },
{ "incgam",2,(void*)incgam0,3,"GGD0,G,p",NULL,NULL,NULL },
{ "incgamc",2,(void*)incgam3,3,"GGp",NULL,NULL,NULL },
{ "intformal",14,(void*)integ,7,"Gn",NULL,NULL,NULL },
{ "intnum",37,(void*)intnum0,9,"V=GGID0,L,p",NULL,NULL,NULL },
{ "isfundamental",1,(void*)gisfundamental,4,"G",NULL,NULL,NULL },
{ "isprime",1,(void*)gisprime,4,"G",NULL,NULL,NULL },
{ "ispseudoprime",1,(void*)gispsp,4,"G",NULL,NULL,NULL },
{ "issquare",1,(void*)gcarreparfait,4,"G",NULL,NULL,NULL },
{ "issquarefree",1,(void*)gissquarefree,4,"G",NULL,NULL,NULL },
{ "kronecker",2,(void*)gkronecker,4,"GG",NULL,NULL,NULL },
{ "label",60,NULL,11,NULL,NULL,NULL,NULL },
{ "lcm",2,(void*)glcm,4,"GG",NULL,NULL,NULL },
{ "length",1,(void*)glength,2,"G",NULL,NULL,NULL },
{ "lex",20,(void*)lexcmp,2,"GGl",NULL,NULL,NULL },
{ "lift",1,(void*)lift,2,"G",NULL,NULL,NULL },
{ "lindep",1,(void*)lindep0,8,"GD0,L,p",NULL,NULL,NULL },
{ "lngamma",1,(void*)glngamma,3,"Gp",NULL,NULL,NULL },
{ "log",1,(void*)log0,3,"GD0,L,p",NULL,NULL,NULL },
{ "matadjoint",1,(void*)adj,8,"G",NULL,NULL,NULL },
{ "matalgtobasis",2,(void*)matalgtobasis,6,"GG",NULL,NULL,NULL },
{ "matbasistoalg",2,(void*)matbasistoalg,6,"GG",NULL,NULL,NULL },
{ "matcompanion",1,(void*)assmat,8,"G",NULL,NULL,NULL },
{ "matdet",1,(void*)det0,8,"GD0,L,",NULL,NULL,NULL },
{ "matdetint",1,(void*)detint,8,"G",NULL,NULL,NULL },
{ "matdiagonal",1,(void*)diagonal,8,"G",NULL,NULL,NULL },
{ "mateigen",1,(void*)eigen,8,"Gp",NULL,NULL,NULL },
{ "mathess",1,(void*)hess,8,"G",NULL,NULL,NULL },
{ "mathilbert",11,(void*)hilb,8,"L",NULL,NULL,NULL },
{ "mathnf",1,(void*)mathnf0,8,"GD0,L,",NULL,NULL,NULL },
{ "mathnfmod",2,(void*)hnfmod,8,"GG",NULL,NULL,NULL },
{ "mathnfmodid",2,(void*)hnfmodid,8,"GG",NULL,NULL,NULL },
{ "matid",11,(void*)idmat,8,"L",NULL,NULL,NULL },
{ "matimage",1,(void*)matimage0,8,"GD0,L,",NULL,NULL,NULL },
{ "matimagecompl",1,(void*)imagecompl,8,"G",NULL,NULL,NULL },
{ "matindexrank",1,(void*)indexrank,8,"G",NULL,NULL,NULL },
{ "matintersect",2,(void*)intersect,8,"GG",NULL,NULL,NULL },
{ "matinverseimage",2,(void*)inverseimage,8,"GG",NULL,NULL,NULL },
{ "matisdiagonal",10,(void*)isdiagonal,8,"Gl",NULL,NULL,NULL },
{ "matker",1,(void*)matker0,8,"GD0,L,p",NULL,NULL,NULL },
{ "matkerint",1,(void*)matkerint0,8,"GD0,L,",NULL,NULL,NULL },
{ "matmuldiagonal",2,(void*)matmuldiagonal,8,"GG",NULL,NULL,NULL },
{ "matmultodiagonal",2,(void*)matmultodiagonal,8,"GG",NULL,NULL,NULL },
{ "matpascal",11,(void*)pasc,8,"L",NULL,NULL,NULL },
{ "matrank",10,(void*)rank,8,"Gl",NULL,NULL,NULL },
{ "matrix",49,(void*)matrice0,8,"GGVVI",NULL,NULL,NULL },
{ "matrixqz",2,(void*)matrixqz0,8,"GG",NULL,NULL,NULL },
{ "matsize",1,(void*)matsize,2,"G",NULL,NULL,NULL },
{ "matsnf",1,(void*)matsnf0,8,"GD0,L,",NULL,NULL,NULL },
{ "matsolve",2,(void*)gauss,8,"GG",NULL,NULL,NULL },
{ "matsolvemod",3,(void*)matsolvemod0,8,"GGGD0,L,",NULL,NULL,NULL },
{ "matsupplement",1,(void*)suppl,8,"Gp",NULL,NULL,NULL },
{ "mattranspose",1,(void*)gtrans,8,"G",NULL,NULL,NULL },
{ "max",2,(void*)gmax,1,"GG",NULL,NULL,NULL },
{ "min",2,(void*)gmin,1,"GG",NULL,NULL,NULL },
{ "modreverse",1,(void*)polymodrecip,6,"G",NULL,NULL,NULL },
{ "moebius",1,(void*)gmu,4,"G",NULL,NULL,NULL },
{ "newtonpoly",2,(void*)newtonpoly,6,"GG",NULL,NULL,NULL },
{ "next",0,(void*)next0,11,"v",NULL,NULL,NULL },
{ "nextprime",1,(void*)gnextprime,4,"G",NULL,NULL,NULL },
{ "nfalgtobasis",2,(void*)algtobasis,6,"GG",NULL,NULL,NULL },
{ "nfbasis",13,(void*)nfbasis0,6,"GD0,L,D0,G,",NULL,NULL,NULL },
{ "nfbasistoalg",2,(void*)basistoalg,6,"GG",NULL,NULL,NULL },
{ "nfdetint",2,(void*)nfdetint,6,"GG",NULL,NULL,NULL },
{ "nfdisc",1,(void*)nfdiscf0,6,"GD0,L,D0,G,",NULL,NULL,NULL },
{ "nfeltdiv",3,(void*)element_div,6,"GGG",NULL,NULL,NULL },
{ "nfeltdiveuc",3,(void*)nfdiveuc,6,"GGG",NULL,NULL,NULL },
{ "nfeltdivrem",3,(void*)nfdivres,6,"GGG",NULL,NULL,NULL },
{ "nfeltmod",3,(void*)nfmod,6,"GGG",NULL,NULL,NULL },
{ "nfeltmul",3,(void*)element_mul,6,"GGG",NULL,NULL,NULL },
{ "nfeltpow",3,(void*)element_pow,6,"GGG",NULL,NULL,NULL },
{ "nfeltreduce",3,(void*)element_reduce,6,"GGG",NULL,NULL,NULL },
{ "nfeltval",30,(void*)element_val,6,"GGGl",NULL,NULL,NULL },
{ "nffactor",30,(void*)nffactor,6,"GG",NULL,NULL,NULL },
{ "nffactormod",30,(void*)nffactormod,6,"GGG",NULL,NULL,NULL },
{ "nfgaloisapply",3,(void*)galoisapply,6,"GGG",NULL,NULL,NULL },
{ "nfgaloisconj",1,(void*)galoisconj0,6,"GD0,L,p",NULL,NULL,NULL },
{ "nfhilbert",10,(void*)nfhilbert0,6,"lGGGD0,G,",NULL,NULL,NULL },
{ "nfhnf",2,(void*)nfhermite,6,"GG",NULL,NULL,NULL },
{ "nfhnfmod",3,(void*)nfhermitemod,6,"GGG",NULL,NULL,NULL },
{ "nfinit",1,(void*)nfinit0,6,"GD0,L,p",NULL,NULL,NULL },
{ "nfisideal",20,(void*)isideal,6,"GGl",NULL,NULL,NULL },
{ "nfisincl",2,(void*)nfisincl0,6,"GGD0,L,",NULL,NULL,NULL },
{ "nfisisom",2,(void*)nfisisom0,6,"GGD0,L,",NULL,NULL,NULL },
{ "nfroots",30,(void*)nfroots,6,"GG",NULL,NULL,NULL },
{ "nfrootsof1",1,(void*)rootsof1,6,"Gp",NULL,NULL,NULL },
{ "nfsnf",2,(void*)nfsmith,6,"GG",NULL,NULL,NULL },
{ "nfsubfields",2,(void*)subfields0,6,"GD0,G,",NULL,NULL,NULL },
{ "norm",1,(void*)gnorm,2,"G",NULL,NULL,NULL },
{ "norml2",1,(void*)gnorml2,2,"G",NULL,NULL,NULL },
{ "numdiv",1,(void*)gnumbdiv,4,"G",NULL,NULL,NULL },
{ "numerator",1,(void*)numer,2,"G",NULL,NULL,NULL },
{ "numtoperm",24,(void*)permute,2,"LG",NULL,NULL,NULL },
{ "omega",1,(void*)gomega,4,"G",NULL,NULL,NULL },
{ "padicappr",2,(void*)apprgen9,7,"GG",NULL,NULL,NULL },
{ "padicprec",20,(void*)padicprec,2,"GGl",NULL,NULL,NULL },
{ "permtonum",1,(void*)permuteInv,2,"G",NULL,NULL,NULL },
{ "polcoeff",21,(void*)truecoeff,2,"GL",NULL,NULL,NULL },
{ "polcompositum",2,(void*)polcompositum0,6,"GGD0,L,",NULL,NULL,NULL },
{ "polcyclo",11,(void*)cyclo,7,"LDx,n,",NULL,NULL,NULL },
{ "poldegree",10,(void*)degree,2,"Gl",NULL,NULL,NULL },
{ "poldisc",1,(void*)discsr,7,"G",NULL,NULL,NULL },
{ "poldiscreduced",1,(void*)reduceddiscsmith,7,"G",NULL,NULL,NULL },
{ "polgalois",1,(void*)galois,6,"Gp",NULL,NULL,NULL },
{ "polinterpolate",31,(void*)polint,7,"GGGF",NULL,NULL,NULL },
{ "polisirreducible",1,(void*)gisirreducible,7,"G",NULL,NULL,NULL },
{ "polkaramul",32,(void*)karamul,7,"GGL",NULL,NULL,NULL },
{ "pollegendre",11,(void*)legendre,7,"LDx,n,",NULL,NULL,NULL },
{ "polrecip",1,(void*)polrecip,7,"G",NULL,NULL,NULL },
{ "polred",1,(void*)polred0,6,"GD0,L,D0,G,p",NULL,NULL,NULL },
{ "polredabs",1,(void*)polredabs0,6,"GD0,L,p",NULL,NULL,NULL },
{ "polredord",1,(void*)ordred,6,"Gp",NULL,NULL,NULL },
{ "polresultant",2,(void*)polresultant0,7,"GGD0,L,",NULL,NULL,NULL },
{ "polroots",1,(void*)roots0,7,"GD0,L,p",NULL,NULL,NULL },
{ "polrootsmod",2,(void*)rootmod0,7,"GGD0,L,",NULL,NULL,NULL },
{ "polrootspadic",32,(void*)rootpadic,7,"GGL",NULL,NULL,NULL },
{ "polsturm",10,(void*)sturm,7,"Gl",NULL,NULL,NULL },
{ "polsturmpart",30,(void*)sturmpart,7,"GGGl",NULL,NULL,NULL },
{ "polsubcyclo",2,(void*)subcyclo,6,"GGDx,n,",NULL,NULL,NULL },
{ "polsylvestermatrix",2,(void*)sylvestermatrix,7,"GGp",NULL,NULL,NULL },
{ "polsym",21,(void*)polsym,7,"GL",NULL,NULL,NULL },
{ "poltchebi",11,(void*)tchebi,7,"LDx,n,",NULL,NULL,NULL },
{ "poltschirnhaus",1,(void*)tschirnhaus,6,"G",NULL,NULL,NULL },
{ "polylog",24,(void*)polylog0,3,"LGD0,L,p",NULL,NULL,NULL },
{ "polzagier",21,(void*)polzag,7,"LL",NULL,NULL,NULL },
{ "precision",1,(void*)precision0,2,"GD0,L,",NULL,NULL,NULL },
{ "precprime",1,(void*)gprecprime,4,"G",NULL,NULL,NULL },
{ "prime",11,(void*)prime,4,"L",NULL,NULL,NULL },
{ "primes",11,(void*)primes,4,"L",NULL,NULL,NULL },
{ "prod",48,(void*)produit,9,"V=GGID1,G,",NULL,NULL,NULL },
{ "prodeuler",37,(void*)prodeuler,9,"V=GGIp",NULL,NULL,NULL },
{ "prodinf",27,(void*)prodinf0,9,"V=GID0,L,p",NULL,NULL,NULL },
{ "psi",1,(void*)gpsi,3,"Gp",NULL,NULL,NULL },
{ "qfbclassno",1,(void*)qfbclassno0,4,"GD0,L,",NULL,NULL,NULL },
{ "qfbcomprealraw",2,(void*)comprealraw,4,"GGp",NULL,NULL,NULL },
{ "qfbhclassno",1,(void*)classno3,4,"G",NULL,NULL,NULL },
{ "qfbnucomp",3,(void*)nucomp,4,"GGG",NULL,NULL,NULL },
{ "qfbnupow",2,(void*)nupow,4,"GG",NULL,NULL,NULL },
{ "qfbpowrealraw",23,(void*)powrealraw,4,"GLp",NULL,NULL,NULL },
{ "qfbprimeform",2,(void*)primeform,4,"GGp",NULL,NULL,NULL },
{ "qfbred",1,(void*)qfbred0,4,"GD0,L,D0,G,D0,G,D0,G,",NULL,NULL,NULL },
{ "qfgaussred",1,(void*)sqred,8,"G",NULL,NULL,NULL },
{ "qfjacobi",1,(void*)jacobi,8,"Gp",NULL,NULL,NULL },
{ "qflll",1,(void*)qflll0,8,"GD0,L,p",NULL,NULL,NULL },
{ "qflllgram",1,(void*)qflllgram0,8,"GD0,L,p",NULL,NULL,NULL },
{ "qfminim",33,(void*)minim0,8,"GGGD0,L,p",NULL,NULL,NULL },
{ "qfperfection",10,(void*)perf,8,"G",NULL,NULL,NULL },
{ "qfsign",1,(void*)signat,8,"G",NULL,NULL,NULL },
{ "quadclassunit",96,(void*)quadclassunit0,4,"GD0,L,D0,G,p",NULL,NULL,NULL },
{ "quaddisc",1,(void*)quaddisc,4,"G",NULL,NULL,NULL },
{ "quadgen",1,(void*)quadgen,2,"G",NULL,NULL,NULL },
{ "quadpoly",1,(void*)quadpoly,2,"G",NULL,NULL,NULL },
{ "quadregulator",1,(void*)gregula,4,"Gp",NULL,NULL,NULL },
{ "quadunit",1,(void*)gfundunit,4,"Gp",NULL,NULL,NULL },
{ "random",0,(void*)genrand,2,"",NULL,NULL,NULL },
{ "real",1,(void*)greal,2,"G",NULL,NULL,NULL },
{ "reorder",0,(void*)reorder,11,"D[],G,",NULL,NULL,NULL },
{ "return",0,(void*)return0,11,"vD0,G,",NULL,NULL,NULL },
{ "rnfalgtobasis",2,(void*)rnfalgtobasis,6,"GG",NULL,NULL,NULL },
{ "rnfbasis",2,(void*)rnfbasis,6,"GG",NULL,NULL,NULL },
{ "rnfbasistoalg",2,(void*)rnfbasistoalg,6,"GG",NULL,NULL,NULL },
{ "rnfcharpoly",14,(void*)rnfcharpoly,6,"GGGDx,n,",NULL,NULL,NULL },
{ "rnfconductor",2,(void*)rnfconductor,6,"GGp",NULL,NULL,NULL },
{ "rnfdedekind",30,(void*)rnfdedekind,6,"GGG",NULL,NULL,NULL },
{ "rnfdet",2,(void*)rnfdet0,6,"GGD0,G,",NULL,NULL,NULL },
{ "rnfdisc",2,(void*)rnfdiscf,6,"GG",NULL,NULL,NULL },
{ "rnfeltabstorel",2,(void*)rnfelementabstorel,6,"GG",NULL,NULL,NULL },
{ "rnfeltdown",2,(void*)rnfelementdown,6,"GG",NULL,NULL,NULL },
{ "rnfeltreltoabs",2,(void*)rnfelementreltoabs,6,"GG",NULL,NULL,NULL },
{ "rnfeltup",2,(void*)rnfelementup,6,"GG",NULL,NULL,NULL },
{ "rnfequation",2,(void*)rnfequationall,6,"GGD0,L,",NULL,NULL,NULL },
{ "rnfhnfbasis",2,(void*)rnfhermitebasis,6,"GG",NULL,NULL,NULL },
{ "rnfidealabstorel",2,(void*)rnfidealabstorel,6,"GG",NULL,NULL,NULL },
{ "rnfidealdown",2,(void*)rnfidealdown,6,"GG",NULL,NULL,NULL },
{ "rnfidealhnf",2,(void*)rnfidealhermite,6,"GG",NULL,NULL,NULL },
{ "rnfidealmul",2,(void*)rnfidealmul,6,"GG",NULL,NULL,NULL },
{ "rnfidealnormabs",2,(void*)rnfidealnormabs,6,"GG",NULL,NULL,NULL },
{ "rnfidealnormrel",2,(void*)rnfidealnormrel,6,"GG",NULL,NULL,NULL },
{ "rnfidealreltoabs",2,(void*)rnfidealreltoabs,6,"GG",NULL,NULL,NULL },
{ "rnfidealtwoelt",2,(void*)rnfidealtwoelement,6,"GG",NULL,NULL,NULL },
{ "rnfidealup",2,(void*)rnfidealup,6,"GG",NULL,NULL,NULL },
{ "rnfinit",2,(void*)rnfinitalg,6,"GGp",NULL,NULL,NULL },
{ "rnfisfree",20,(void*)rnfisfree,6,"GGl",NULL,NULL,NULL },
{ "rnfisnorm",3,(void*)rnfisnorm,6,"GGGD1,L,p",NULL,NULL,NULL },
{ "rnflllgram",3,(void*)rnflllgram,6,"GGGp",NULL,NULL,NULL },
{ "rnfnormgroup",2,(void*)rnfnormgroup,6,"GG",NULL,NULL,NULL },
{ "rnfpolred",2,(void*)rnfpolred,6,"GGp",NULL,NULL,NULL },
{ "rnfpseudobasis",2,(void*)rnfpseudobasis,6,"GG",NULL,NULL,NULL },
{ "rnfsteinitz",2,(void*)rnfsteinitz,6,"GG",NULL,NULL,NULL },
{ "round",1,(void*)round0,2,"GD0,L,",NULL,NULL,NULL },
{ "rounderror",10,(void*)rounderror,2,"Gl",NULL,NULL,NULL },
{ "serconvol",2,(void*)convol,7,"GG",NULL,NULL,NULL },
{ "serlaplace",1,(void*)laplace,7,"G",NULL,NULL,NULL },
{ "serreverse",1,(void*)recip,7,"G",NULL,NULL,NULL },
{ "setintersect",2,(void*)setintersect,8,"GG",NULL,NULL,NULL },
{ "setisset",10,(void*)isvecset,8,"Gl",NULL,NULL,NULL },
{ "setminus",2,(void*)setminus,8,"GG",NULL,NULL,NULL },
{ "setrand",11,(void*)setrand,11,"Lp",NULL,NULL,NULL },
{ "setsearch",20,(void*)setsearch,8,"GGl",NULL,NULL,NULL },
{ "setunion",2,(void*)setunion,8,"GG",NULL,NULL,NULL },
{ "shift",21,(void*)gshift,1,"GL",NULL,NULL,NULL },
{ "shiftmul",21,(void*)gmul2n,1,"GL",NULL,NULL,NULL },
{ "sigma",1,(void*)gsumdivk,4,"GD1,L,",NULL,NULL,NULL },
{ "sign",10,(void*)gsigne,1,"Gl",NULL,NULL,NULL },
{ "simplify",1,(void*)simplify,2,"G",NULL,NULL,NULL },
{ "sin",1,(void*)gsin,3,"Gp",NULL,NULL,NULL },
{ "sinh",1,(void*)gsh,3,"Gp",NULL,NULL,NULL },
{ "sizebyte",10,(void*)taille2,2,"Gl",NULL,NULL,NULL },
{ "sizedigit",10,(void*)gsize,2,"Gl",NULL,NULL,NULL },
{ "solve",37,(void*)zbrent,9,"V=GGIp",NULL,NULL,NULL },
{ "sqr",1,(void*)gsqr,3,"G",NULL,NULL,NULL },
{ "sqrt",1,(void*)gsqrt,3,"Gp",NULL,NULL,NULL },
{ "sqrtint",1,(void*)racine,4,"Gp",NULL,NULL,NULL },
{ "subst",26,(void*)gsubst,7,"GnG",NULL,NULL,NULL },
{ "sum",48,(void*)somme,9,"V=GGID0,G,p",NULL,NULL,NULL },
{ "sumalt",27,(void*)sumalt0,9,"V=GID0,L,p",NULL,NULL,NULL },
{ "sumdiv",22,(void*)divsum0,9,"GVI",NULL,NULL,NULL },
{ "suminf",27,(void*)suminf,9,"V=GIp",NULL,NULL,NULL },
{ "sumpos",27,(void*)sumpos0,9,"V=GID0,L,p",NULL,NULL,NULL },
{ "tan",1,(void*)gtan,3,"Gp",NULL,NULL,NULL },
{ "tanh",1,(void*)gth,3,"Gp",NULL,NULL,NULL },
{ "taylor",12,(void*)tayl,7,"GnP",NULL,NULL,NULL },
{ "teichmuller",1,(void*)teich,3,"Gp",NULL,NULL,NULL },
{ "theta",2,(void*)theta,3,"GGp",NULL,NULL,NULL },
{ "thetanullk",21,(void*)thetanullk,3,"GL",NULL,NULL,NULL },
{ "thue",2,(void*)thue,7,"GGD0,G,",NULL,NULL,NULL },
{ "thueinit",2,(void*)thueinit,7,"GD0,L,p",NULL,NULL,NULL },
{ "trace",1,(void*)trace,8,"Gp",NULL,NULL,NULL },
{ "truncate",1,(void*)trunc0,2,"GD0,L,",NULL,NULL,NULL },
{ "until",82,NULL,11,NULL,NULL,NULL,NULL },
{ "valuation",20,(void*)ggval,2,"GGl",NULL,NULL,NULL },
{ "variable",1,(void*)gpolvar,2,"G",NULL,NULL,NULL },
{ "vecconcat",2,(void*)concat,8,"GG",NULL,NULL,NULL },
{ "vecextract",2,(void*)extract0,8,"GGD-1,G,",NULL,NULL,NULL },
{ "vecindexsort",1,(void*)indexsort,8,"G",NULL,NULL,NULL },
{ "veclexsort",1,(void*)lexsort,8,"G",NULL,NULL,NULL },
{ "vecmax",1,(void*)vecmax,2,"Gp",NULL,NULL,NULL },
{ "vecmin",1,(void*)vecmin,2,"Gp",NULL,NULL,NULL },
{ "vecsort",2,(void*)vecsort0,8,"GD0,G,p",NULL,NULL,NULL },
{ "vector",22,(void*)vector0,8,"GVI",NULL,NULL,NULL },
{ "vectorv",22,(void*)vvector0,8,"GVI",NULL,NULL,NULL },
{ "weber",1,(void*)weber0,3,"GD0,L,p",NULL,NULL,NULL },
{ "while",81,NULL,11,NULL,NULL,NULL,NULL },
{ "zeta",1,(void*)gzeta,3,"Gp",NULL,NULL,NULL },
{ "zetak",2,(void*)gzetakall,6,"GGD0,L,p",NULL,NULL,NULL },
{ "zetakinit",1,(void*)initzeta,6,"Gp",NULL,NULL,NULL },
{ "znorder",1,(void*)order,4,"G",NULL,NULL,NULL },
{ "znprimroot",1,(void*)ggener,4,"G",NULL,NULL,NULL },
{ "znstar",1,(void*)znstar,4,"Gp",NULL,NULL,NULL },

/* DO NOT REMOVE THIS BLANK LINE: chname & helpsynchro depend on it */
{ NULL,0,NULL,0,NULL,NULL,NULL,NULL } /* sentinel */
};

static GEN
seq(void)
{
  const long av=avma, lim=(avma+bot)>>1, oldtetpil=analyseurtetpil;
  GEN res=gnil;

  check_break_status(DOCLEAR,NULL);
  for(;;)
  {
    while (separe(*analyseur)) analyseur++;
    if (!*analyseur || *analyseur == ')' || *analyseur == ',') return res;
    res = expr();
    if (check_break_status(NOACTION,NULL)) return NULL;
    if (!separe(*analyseur)) return res;
    if (avma<lim)
    {
      if(DEBUGMEM>1) fprintferr("Collecting garbage in `seq'\n");
      if (is_universal_constant(res))
	analyseurtetpil = avma = av;
      else
	{ res=gerepileupto(av,forcecopy(res)); analyseurtetpil = oldtetpil; }
    }
  }
}

static GEN
expr(void)
{
  PFGEN func[] = { NULL,NULL,NULL,NULL };
  long niveau = 3;
  GEN aux,e,e1,e2,e3;

  check_break_status(DOCLEAR,NULL);
  e1=e2=e3=NULL;
  for(;;)
    switch(niveau)
    {
      case 3: aux=facteur();
	if (check_break_status(NOACTION,NULL)) return NULL;
	if (func[3]) 
	{
	  analyseurtetpil = avma;
	  e3 = ((GEN (*)(GEN,GEN))func[3])(e3,aux);
	}
	else
	  e3=aux;
	switch(*analyseur)
	{
	  case '*': analyseur++; func[3] = (PFGEN)&gmul; break;
	  case '/': analyseur++; func[3] = (PFGEN)&gdiv; break;
	  case '\\': analyseur++;
	    if (*analyseur == '/') { analyseur++; func[3]=(PFGEN)&gdivround; }
	    else func[3] = (PFGEN)&gdivent;
	    break;
	  case '%': analyseur++; func[3] = (PFGEN)&gmod; break;

	  case '<': case '>':
	    if (analyseur[1]==*analyseur)
	    {
	      func[3] = (*analyseur == '<')? (PFGEN)&gshiftplus
	                                   : (PFGEN)&gshiftminus;
	      analyseur +=2; break;
	    }
	    /* fall through */
	  default: niveau--; func[3]=NULL;
	}
	break;
      case 2:
	if (!e3) { niveau++; break; }
	if (func[2])
	{
	  analyseurtetpil = avma;
	  e2 = ((GEN (*)(GEN,GEN))func[2])(e2,e3);
	}
	else
	  e2=e3;
	e3=NULL;
	switch(*analyseur)
	{
	  case '+': analyseur++; func[2]=(PFGEN)&gadd; niveau++; break;
	  case '-': analyseur++; func[2]=(PFGEN)&gsub; niveau++; break;
	  default: niveau--; func[2]=NULL;
	}
	break;
      case 1:
	if (!e2) { niveau++; break; }
	if (func[1])
	{
	  analyseurtetpil = avma;
	  e1 = ((GEN (*)(GEN,GEN))func[1])(e1,e2);
	}
	else
	  e1=e2;
	e2=NULL;
	switch(*analyseur)
	{
	  case '<': analyseur++;
	    switch(*analyseur)
	    {
	      case '=': analyseur++; func[1]=(PFGEN)&gle; break;
	      case '>': analyseur++; func[1]=(PFGEN)&gne; break;
	      default : func[1]=(PFGEN)&glt;
	    }
	    niveau++; break;
	  case '>': analyseur++;
	    if (*analyseur == '=') { analyseur++; func[1]=(PFGEN)&gge; }
	    else func[1]=(PFGEN)&ggt;
	    niveau++; break;
	  case '=':
	    if (analyseur[1] == '=')
	      { analyseur+=2; func[1]=(PFGEN)&geq; niveau++; }
	    break;
	  case '!':
	    if (analyseur[1] == '=')
	      { analyseur+=2; func[1]=(PFGEN)&gne; niveau++; }
	    break;
	  default: niveau--; func[1]=NULL;
	}
	break;
      case 0:
	if (!e1) { niveau++; break; }
	if (func[0])
	  e=gcmp0(e1) ? gzero: gun;
	else
	  e=e1;
	e1=NULL;
	switch(*analyseur)
	{
	  case '&':
	    if (*++analyseur == '&') analyseur++;
	    if (gcmp0(e)) { skipexpr(); return gzero; }
	    func[0]=(PFGEN)1; niveau++;
	    break;
	  case '|':
	    if (*++analyseur == '|') analyseur++;
	    if (!gcmp0(e)) { skipexpr(); return gun; }
	    func[0]=(PFGEN)1; niveau++;
	    break;
	  default: return e;
	}
    }
}

/********************************************************************/
/**                                                                **/
/**                        CHECK FUNCTIONS                         **/
/**                                                                **/
/********************************************************************/

static void
match(char c)
{
  if (*analyseur != c)
  {
    char str[64];

    if (check_new_fun) err_new_fun();
    sprintf(str,"expected character: '%c'\n  ***   instead of: ",c);
    err(talker2,str,analyseur,mark.start);
  }
  analyseur++;
}

static long
readlong()
{
  long m;
  const long av = avma;
  const char *old = analyseur;
  const GEN arg = expr();

  if (check_break_status(NOACTION,NULL))
    err(talker,"break not allowed here");

  if (typ(arg) != t_INT) err(caseer,old,mark.start);
  m = itos(arg); avma=av;
  return m;
}

static long
check_array_index(GEN p)
{
  const char *old = analyseur;
  const long c = readlong();

  if (c < 1 || c >= lg(p)) err(arrayer1,old,mark.start);
  return c;
}

static int
readvar()
{
  const char *old = analyseur;
  const GEN x = expr();

  if (check_break_status(NOACTION,NULL))
    err(talker,"break not allowed here");
  if (typ(x) != t_POL) err(numvarer,old,mark.start);
  if (lgef(x) != 4) err(numvarer,old,mark.start);
  if ( !gcmp0((GEN)x[2]) || !gcmp1((GEN)x[3]) ) err(numvarer,old,mark.start);
  return varn(x);
}

/* allright !=0 means function was called without () */
/* matchcomma = *l should be 0 or 1 */
static int
do_switch(int allright, int matchcomma)
{
  if (!*analyseur || *analyseur == ')' || separe(*analyseur)) return 1;
  if (*analyseur == ',') /* we just read an arg, or first arg */
  {
    if (allright) return 1; /* arg to an external function */
    if (!matchcomma && analyseur[-1] == '(') return 1; /* first arg */
    if (analyseur[1] == ',' || analyseur[1] == ')')
      { analyseur++; return 1; }
  }
  return 0;
}

/********************************************************************/
/**                                                                **/
/**                          READ FUNCTIONS                        **/
/**                                                                **/
/********************************************************************/

static GEN
facteur(void)
{
  const char *old = analyseur;
  GEN x,p1;
  int plus=1;

  switch(*analyseur)
  {
    case '+': analyseur++; break;
    case '-': plus=0; analyseur++; break;
  }
  x=truc();
  if (check_break_status(NOACTION,NULL))
    return NULL;

  for(;;)
    switch(*analyseur)
    {
      case '.':
	analyseur++; x = read_member(x); break;
      case '^':
	analyseur++; p1=facteur(); analyseurtetpil=avma;
	if (check_break_status(NOACTION,NULL))
	  err(talker,"break not allowed after ^");
	x=gpui(x,p1,prec); break;
      case '_':
	analyseur++; analyseurtetpil=avma; x=gconj(x); break;
      case '\'':
	analyseur++; analyseurtetpil=avma; x=deriv(x,gvar9(x)); break;
      case '~':
	analyseur++; analyseurtetpil=avma; x=gtrans(x); break;
      case '[':
        x=matrix_block(x,NULL); break;
      case '!':
	if (analyseur[1] != '=')
	{
	  if (typ(x) != t_INT) err(caseer,old,mark.start);
	  analyseur++; analyseurtetpil=avma;
	  x=mpfact(itos(x)); break;
	} /* Fall through */

      default:
        if (plus || x==gnil) return x;
	analyseurtetpil = avma; return gneg(x);
    }
}

static GEN
truc(void)
{
  long i,j, n=0, p=0, m=1;
  GEN *table,p1;
  char *old;

  if (*analyseur == '!')
  {
    analyseur++;
    i = gcmp0(truc());
    if (check_break_status(NOACTION,NULL))
      err(talker,"break not allowed after !");
    return i ? gun : gzero;
  }
  if (isalpha(*analyseur)) return identifier();

  analyseurtetpil = avma;
  if (isdigit(*analyseur) || *analyseur == '.') return constante();
  switch(*analyseur++)
  {
    case '(': p1=expr(); match(')'); return p1;

    case '[':
      if (*analyseur == ';' && analyseur[1] == ']')
	{ analyseur+=2; return cgetg(1,t_MAT); }
      table = (GEN*) newbloc(strlen(analyseur));
      if (*analyseur != ']') table[++n]=expr();
      while (*analyseur == ',') { analyseur++; table[++n]=expr(); }
      if (check_break_status(NOACTION,NULL))
	err(talker,"break not allowed here");
      switch (*analyseur++)
      {
	case ']':
	  analyseurtetpil = avma;
	  p1=cgetg(n+1,t_VEC);
	  for (i=1; i<=n; i++)
	    p1[i]=lcopy(table[i]);
	  break;

	case ';':
	  m = n;
	  do 
	    table[++n]=expr();
	  while (*analyseur++!=']');
	  if (check_break_status(NOACTION,NULL))
	    err(talker,"break not allowed here");
	  p=n/m; analyseurtetpil=avma;
	  p1=cgetg(m+1,t_MAT);
	  for (j=1; j<=m; j++)
	  {
	    p1[j]=lgetg(p+1,t_COL);
	    for (i=1; i<=p; i++)
	      coeff(p1,i,j) = lcopy(table[(i-1)*m+j]);
	  }
	  break;

	default:
          /* can only occur in library mode */
          err(talker,"incorrect vector or matrix");
          return NULL; /* not reached */
      }
      killbloc((GEN)table); return p1;

    case '%':
      old=analyseur-1; p=0;
      if (! gp_history_fun)
	err(talker2,"history not available in library mode: ",old,mark.start);
      while ( *analyseur == '`') { analyseur++; p++; }
      return p ? gp_history_fun(p         ,1,old,mark.start)
               : gp_history_fun(number(&n),0,old,mark.start);
  }
  err(caracer1,analyseur-1,mark.start);
  return NULL; /* not reached */
}

#define can_repeat_op(c) ((c) == '+' || (c) == '-') /* valid x opop, eg x++*/

static GEN
matrix_block(GEN p, entree *ep)
{
  long tx,full_col,full_row,c,e;
  char *old;
  GEN res, p1;

  tx = full_col = full_row = 0; p1 = p;
  while (*analyseur == '[')
  {
    p = p1; tx=typ(p); full_col = full_row = 0;
    if (! is_matvec_t(tx)) err(caracer1,analyseur,mark.start);
    analyseur++;
    if (tx != t_MAT) c=check_array_index(p); /* vector */
    else
    {
      /* matrix */
      if (lg(p)==1) err(arrayer1,analyseur,mark.start); /* 0x0 matrix */
      if (*analyseur==',') /* whole column */
      {
	full_col=1; analyseur++; c=check_array_index(p);
      }
      else
      {
	e=check_array_index((GEN) p[1]); match(',');
	if (*analyseur == ']') full_row=1; /* whole row */
	else c=check_array_index(p);
      }
    }
    match(']');

    if (tx == t_MAT)
    {
      if (full_col) p1 = (GEN) p[c];
      else if (full_row)
      {
	if (*analyseur != ']') break;

	res = cgetg(lg(p),t_VEC);
	for (c=1; c<lg(p); c++) res[c] = coeff(p,e,c);
	p1=res;
      }
      else p1 = gcoeff(p,e,c);
    }
    else p1 = (GEN) p[c];
  }
  analyseurtetpil = avma; old = analyseur;

  if (*analyseur == '=') /* affectation or equality test */
  {
      if (analyseur[1] == '=')  /* test */
      { 
	if (!full_row) return p1;
	res = cgetg(lg(p),t_VEC);
	for (c=1; c<lg(p); c++) res[c] = coeff(p,e,c);
	return res;
      }
      analyseur++; old=analyseur; res=expr(); /* affectation */
  }
  else if (can_repeat_op(*analyseur) && *analyseur == analyseur[1])
  {
    /* a++, a-- */
    if (full_row) err(gadder5);
    if (typ(p1)==t_INT)
      res = (*analyseur == '+')? addis(p1,1): subis(p1,1);
    else
      res = (*analyseur == '+')? gadd(p1,gun): gsub(p1,gun);
    analyseur += 2;
  }
  else
  {
    GEN (*f)(GEN,GEN) = NULL;

    if (full_row)
    {
      res = cgetg(lg(p),t_VEC);
      for (c=1; c<lg(p); c++) res[c] = coeff(p,e,c);
      p1=res;
    }

    if (!*analyseur) return (ep && !full_row) ? p1: gcopy(p1);

    /* op= constructs ? */

    if (analyseur[1] != '=')
    {
      switch(*analyseur)
      {
	case '>' :
	  if (analyseur[1]=='>' && analyseur[2] == '=') f = &gshiftminus;
	  break;
	case '<' :
	  if (analyseur[1]=='<' && analyseur[2] == '=') f = &gshiftplus;
	  break;
	case '\\':
	  if (analyseur[1]=='/' && analyseur[2] == '=') f = &gdivround;
	  break;
      }
      if (!f) return (ep && !full_row) ? p1: gcopy(p1);
      analyseur+=3;
    }
    else
    {
      switch(*analyseur)
      {
	case '+' : f = &gadd     ; break;
	case '-' : f = &gsub     ; break;
	case '*' : f = &gmul     ; break;
	case '/' : f = &gdiv     ; break; 
	case '\\': f = &gdivent; break; 
	case '%' : f = &gmod     ; break; 
	default: return (ep && !full_row) ? p1: gcopy(p1);
      }
      analyseur+=2; 
    }
    old=analyseur; res = f(p1,expr());
  }

  /* affectation */

  if (!ep) err(caracer1,analyseur,mark.start);
  if (check_break_status(NOACTION,NULL))
    err(talker,"break not allowed here");

  if (tx == t_MAT)
  {
    if (full_row) /* whole row; e = row number */
    {
      if (typ(res) != t_VEC || lg(res) != lg(p))
	err(caseer2,old,mark.start);

      for (c=1; c<lg(p); c++) coeff(p,e,c) = res[c];
    }
    else if (full_col) /* whole column; c = column number */
    {
      if (typ(res) != t_COL || lg(res) != lg((GEN) p[1]))
	err(caseer2,old,mark.start);

      p[c] = (long) res;
    }
    else /* single element: e=row, c=column */
    {
      if (isclone(p1)) killbloc(p1);
      res = gclone(res); setisclone(res);
      coeff(p,e,c) = (long) res; return res;
    }
  }
  else if (tx) /* vector */
  {
    if (isclone(p1)) killbloc(p1);
    res = gclone(res); setisclone(res);
    p[c] = (long) res; return res;
  }
  else /* simple variable */
  {
    changevalue(ep,(GEN)res);
    return (GEN) ep->value;
  }
}

/*
 *  Read a "string" from *src. Format then Copy it (without \0) at s.
 *  Return a pointer to end of string.
 *
 *  if flag & 1 add the ending '\0'
 *  if flag & 2 set src to the end of string
 */
char *
readstring(char **src, char *s, char *lim, long flag)
{
  const int do_zero = flag & 1;
  const int do_set = flag & 2;
  char *old=analyseur;

  analyseur=*src; match('"');
  while (*analyseur)
  {
    while (*analyseur == '\\')
    {
      switch(*++analyseur)
      {
	case 'e':  *s='\033'; break; /* escape */
	case 'n':  *s='\n'; break;
	case 't':  *s='\t'; break;
	default:   *s=*analyseur;
      }
      analyseur++; s++;
    }
    if ( *analyseur == '"' || s > lim ) break;
    *s++ = *analyseur++;
  }
  if (s > lim) err(talker,"string too long");
  match('"');
  lim=analyseur; analyseur=old; /* in case *src = analyseur... */
  if (do_zero) *s++=0; if (do_set) *src=lim;
  return s;
}

static GEN
call_fun(entree **p)
{
  GEN x, res = lisseq((char *)p);

  if (check_break_status(DOFUNCTION, &x)) res = x;
  else if (! is_universal_constant(res))
  {
    /* make result safe */
    analyseurtetpil=avma;
    res = forcecopy(res);     
  }
  while (*--p) killvalue(*p); /* pop ancient values for formal parameters */
  return res;
}

/* test if current function existed in previous version, before raising
 * an error
 */
static void
err_new_fun()
{
  char *s = NULL, str[128], msg[128];
  long len=0, hash;

  if (check_new_fun)
  { 
    if (check_new_fun != NOT_CREATED_YET) 
    {
      s = strcpy(str,check_new_fun->name);
      kill0(check_new_fun);
    }
    check_new_fun=NULL;
  }
  if (compatible != NONE) return;

  if (!s)
  {
    s = mark.identifier;
    while (isalnum(s[len])) len++;
    s = strncpy(str,s,len); str[len]=0;
  }

  hash = hashvalue(&s,functions_tblsz,1);
  if (!findentry(hash,str,s,funct_old_hash)) return;
  sprintf(msg,"obsolete function; see whatnow(%s): \n  ",str);
  err(talker2, msg, mark.identifier, mark.start);
}

static entree *
do_alias(entree *ep)
{
  while (ep->valence == EpALIAS) ep = (entree *) ep->value;
  return ep;
}

#define TOTAL_STRING_ARG 4096 /* Total length of string args */
#define match_comma() if (matchcomma) match(','); else matchcomma = 1

static GEN
identifier(void)
{
  long m,i,av,matchcomma;
  char *ch1, *old;
  entree *ep;
  GEN res;

  mark.identifier = analyseur; ep = entry();
  if (EpVALENCE(ep)==EpVAR) /* variable */
  {
    if (*analyseur == ')' || *analyseur ==',')
      return ep->value; /* optimized for simple variables */

    return matrix_block((GEN) ep->value,ep);
  }

  ep=do_alias(ep); matchcomma=0;
  if (ep->code)
  {
    static char buffer[TOTAL_STRING_ARG], *limit = buffer+TOTAL_STRING_ARG-2;
    int ret = RET_GEN, allright=0;
    long fake;
    char *old_fun = current_function, *bp = buffer, *s = ep->code;
    char *sinit, *oldanalyseur = NULL;
    GEN *fakepGEN = NULL, fakeGEN, argvec[9];
    void *call = ep->value;

    current_function=ep->name;
    if (*analyseur != '(')
    {
      /* if no mandatory argument, no () needed */
      if (EpVALENCE(ep) != 0)  match('('); /* error */

      if (!*s || (!s[1] && *s == 'p'))
      {
	res = ((GEN (*)(long))ep->value)(prec);
	current_function=old_fun; return res;
      }
      allright=1; /* no arg was given, but valence is ok */
    } else analyseur++;

    i=0;
    /* Optimized for G and p. */
    while (*s == 'G')
    {
      match_comma(); argvec[i++] = expr(); s++;
    }
    if (*s == 'p') { argvec[i++] = (GEN) prec; s++; }

    while (*s)
    {
      switch (*s++)
      {
	case 'G': /* GEN */
	  match_comma(); argvec[i++] = expr(); break;

	case 'L': /* long */
	  match_comma(); argvec[i++] = (GEN) readlong(); break;

	case 'n': /* var number */
	  match_comma(); argvec[i++] = (GEN) readvar(); break;

	case 'V': /* variable */
	case 'S': /* symbol */
	  match_comma(); mark.symbol=analyseur; /* used by kill0() only */
	  argvec[i++] = (GEN) entry();
	  break;

	case  'I': /* Input position */
	  match_comma();
	  argvec[i++] = (GEN) analyseur;
	  skipseq(); break;

	case 'r': /* raw */
	  match_comma(); mark.raw = analyseur; sinit = bp;
	  while (*analyseur)
	  {
	    if (bp > limit) break;
	    while (*analyseur == '"')
	      bp = readstring(&analyseur,bp,limit,2);
	    if (*analyseur == ',' || *analyseur == ')') break;
	    *bp++ = *analyseur++;
	  }
	  if (bp > limit) err(talker,"arg. string too long ");
	  *bp++ = 0; argvec[i++] = (GEN) sinit;
	  break;

	case 's': /* expanded string; empty arg yields "" */
	  match_comma(); sinit = bp;
	  if (*s == '*') /* any number of string objects */
	  {
	    long nb=0; /* nb = number of connected components */
	    int last_is_string=0;

	    s++; bp += sizeof(long); /* to store nb */
	    while (*analyseur)
	    {
	      if (bp > limit) break; 
	      while (*analyseur == '"')
	      {
		bp = readstring(&analyseur,bp,limit,2);
		last_is_string = 1;
	      }
	      if (*analyseur == ')' || *analyseur == ';') break;
	      if (*analyseur == ',') analyseur++;
	      else
	      {
	        GEN res=expr();
		*bp++=0; if (nb || last_is_string) *bp++ = 0;

		/* with *((GEN *)bp) = res, alignment is wrong */
	        memcpy(bp,&res,sizeof(GEN)); bp += sizeof(GEN);
		nb++; if (last_is_string) nb++;
		last_is_string=0;
	      }
	    }
	    if (last_is_string) nb++;
	    memcpy(sinit,&nb,sizeof(long));
	  }
	  else
	  {
	    while (*analyseur)
	    {
	      if (bp > limit) break;
	      if (*analyseur == '"')
		bp = readstring(&analyseur,bp,limit,2);
	      else if (*analyseur == ',' || *analyseur == ')') break;
	      else
	      { /* expand string */
		char *tmp = GENtostring(expr()), *tbp = bp;
		long len = strlen(tmp);

		bp += len; if (bp>limit) break;
		memcpy(tbp,tmp,len*sizeof(long)); /* ignore trailing \0 */
		free(tmp);
	      }
	    }
	  }
	  if (bp > limit) err(talker,"string too long");
	  *bp++ = 0; analyseurtetpil = avma;
	  argvec[i++] = (GEN) sinit;
	  break;

	case 'p': /* precision */
	  argvec[i++] = (GEN) prec; break;

	case '=':
	  match('='); matchcomma = 0; break;

	case 'D': /* Has a default value */
	  if (do_switch(allright,matchcomma))
	  {
	    oldanalyseur = analyseur;
	    analyseur = s; matchcomma = 0;
	  }
	  while (*s++ != ',');
	  break;

	 case 'P': /* series precision */
	   argvec[i++] = (GEN) precdl; break;

	 case 'f': /* Fake *long argument */
	   argvec[i++] = (GEN) &fake; break;

	 case 'F': /* Fake *GEN argument */
	   argvec[i++] = (GEN) &fakeGEN; fakepGEN = &fakeGEN; break;

	 case 'x': /* Foreign function */
	   argvec[i++] = (GEN) ep; call = foreignHandler; break;

	 case 'l': /* Return long */
	   ret = RET_INT; break;

	 case 'v': /* Return void */
	   ret = RET_VOID; break;

	 case ',': /* Clean up default */
	   if (oldanalyseur)
	   {
	     analyseur = oldanalyseur;
	     oldanalyseur = NULL;
	     matchcomma=1;
	   }
	   break;
	 default: err(talker,"Unknown code. Please report");
      }
    }
    if (check_break_status(NOACTION,NULL))
      err(talker,"break not allowed here");

    for ( ;i<9;i++) argvec[i]=NULL; /* unitialized read otherwise */
    switch (ret)
    {
      case RET_GEN:
	analyseurtetpil = avma;
	res = ((PFGEN)call)(argvec[0], argvec[1], argvec[2], argvec[3],
	          argvec[4], argvec[5], argvec[6], argvec[7], argvec[8]);
	break;

      case RET_INT:
	m = ((long (*)(ANYARG))call)(argvec[0], argvec[1], argvec[2], argvec[3],
		  argvec[4], argvec[5], argvec[6], argvec[7], argvec[8]);
	analyseurtetpil = avma; res=stoi(m); break;

      case RET_VOID:
	((void (*)(ANYARG))call)(argvec[0], argvec[1], argvec[2], argvec[3],
	          argvec[4], argvec[5], argvec[6], argvec[7], argvec[8]);
	analyseurtetpil = avma; res=gnil; break;
    }
    if (fakepGEN) cgiv(fakeGEN);
    if (!allright) match(')');
    current_function=old_fun;
    return res;
  }

  if (EpVALENCE(ep) < EpUSER) /* fonctions predefinies */
  {
    res = gnil; match('(');
    switch(EpVALENCE(ep))
    {
      static long yatileugoto;

      case 50: /* O */
        res=truc(); m=1;
	if (check_break_status(NOACTION,NULL))
	  err(talker,"break not allowed in O()");
	if (*analyseur=='^') { analyseur++; m = readlong(); }
	analyseurtetpil = avma;
	res = ggrando(res,m); break;

      case 60: /* label */
        skipexpr(); break;

      case 61: /* goto */
      {
	old=analyseur; m=readlong();
	if (m>=100 || m<0 || !labellist[m]) err(labeler,old,mark.start);
	analyseur=labellist[m]; yatileugoto=1; break;
      }

      case 80: /* if then else */
        av=avma; m=gcmp0(expr()); 
	if (check_break_status(NOACTION,NULL))
	  err(talker,"break not allowed in test expression");
        analyseurtetpil = avma = av; match(',');
	if (m) /* false */
	{
	  skipseq();
	  if (*analyseur == ')') res=gnil;
	  else { match(','); res = seq(); }
	  if (check_break_status(NOACTION,NULL))
	    skipseq();
	}
	else /* true */
	{
	  yatileugoto=0; res = seq();
	  if (check_break_status(NOACTION,NULL))
	    skipseq();
	  if (!yatileugoto)
	    if (*analyseur != ')')
	      { match(','); skipseq(); }
	} 
	break;

      case 81: /* while do */
        analyseurtetpil = av = avma; ch1 = analyseur;
	for(;;)
	{
	  if (gcmp0(expr()))
	  {
	    match(','); skipseq(); break;
	  }
	  avma = av; match(',');
	  yatileugoto=0; seq();
	  if (yatileugoto) break;
	  if (check_break_status(DOLOOP,NULL))
	    { skipseq(); break; }
	  analyseur = ch1;
	}
	analyseurtetpil = avma = av;
	break;

      case 82: /* repeat until*/
        analyseurtetpil = av = avma; ch1 = analyseur;
	skipexpr();
	for(;;)
	{
	  avma = av; match(',');
	  yatileugoto=0; seq();
	  if (yatileugoto) break;
	  if (check_break_status(DOLOOP,NULL))
	    { skipseq(); break; }

	  analyseur = ch1;
	  if (!gcmp0(expr()))
	  {
	    match(','); skipseq(); break;
	  }
	}
	analyseurtetpil = avma = av;
	break;

      default: err(valencer1);
    }
    match(')'); return res;
  }

  switch (EpVALENCE(ep))
  {
    GEN *argdef, *tmparg; /* argdef = the default args (NULL terminated) */
    entree **p;
    char *ch2;
    long nparam;

    case EpUSER: /* user-defined functions */
      argdef = ep->args; p = (entree**) ep->value;
      p++; /* skip the initial NULL, p points at first formal parameter */

      if (*analyseur != '(') /* no args */
      {
	if ( *analyseur != '='  ||  analyseur[1] == '=' )
	{
	  while (*argdef) newvalue(*p++,*argdef++);
	  return call_fun(p);
	}
	match('('); /* ==> error */
      }
      ch1 = analyseur; analyseur++; matchcomma=0;

      nparam = EpNPARAM(ep); 
      tmparg = (GEN*) xmalloc((1+nparam)*sizeof(GEN));
      for (i=0;i<nparam;i++)
	if (do_switch(0,matchcomma))
	  { matchcomma=1 ; tmparg[i] = argdef[i]; } /* defaut arg */
	else
	  { match_comma(); tmparg[i] = expr(); }  /* user suplied */

      /* can't do it immediately. args may use former parameter values */
      for (i=0;i<nparam;i++) newvalue(p[i],tmparg[i]);
      free(tmparg);

      if (*analyseur == ')')
	if (analyseur[1] != '=' || analyseur[2] == '=')
	{ 
	  /* p + nparam points to function body */
	  analyseur++; return call_fun(p + nparam);
	}

      /* Redefine function. Can't occur if we went through skipidentifier. */
      p += nparam;
      while (*--p) killvalue(*p);
      if (*analyseur != ',' && *analyseur != ')') skipexpr();
      while (*analyseur == ',') { analyseur++; skipexpr(); }
      match(')');

      if ( *analyseur != '='  ||  analyseur[1] == '=' )
	err(nparamer1,mark.identifier,mark.start);

      matchcomma=0; analyseur = ch1; analyseurtetpil = avma;
      free_args(ep->args); killbloc((GEN)ep->value);

    /* Fall through */

    case EpNEW: /* new function */
    {
      long len;
      entree **newfun;

      nparam=0; check_new_fun = ep;

	/* checking arguments */
      match('('); ch1 = analyseur;
      while (*analyseur != ')')
      {
	old=analyseur; nparam++; match_comma();
	if (!isalpha(*analyseur))
        {
	  err_new_fun();
	  err(paramer1, mark.identifier, mark.start);
	}

	if (EpVALENCE(skipentry()) != EpVAR)
        {
	  err_new_fun();
	  err(paramer1,old,mark.start);
	}
	if (*analyseur == '=') { analyseur++; skipexpr(); }
      }
      match(')'); match('=');

	/* checking function definition */
      skipping_fun_def++;
      ch2 = analyseur; skipseq(); len = analyseur-ch2;
      skipping_fun_def--;

      /* function  is ok. record it */
      ch2 = analyseur;
      newfun = (entree**) newbloc( nparam + (len>>TWOPOTBYTES_IN_LONG) + 4);
      newfun[-1] = (entree *)ep->value;
      ep->value = (void *)newfun;
      newfun++; /* this bloc is no GEN, leave the first cell alone ( = 0) */

	/* record default args */
      ep->args = argdef = (GEN*) xmalloc((nparam+1)*sizeof(GEN));
      analyseur = ch1; matchcomma=0;
      for (i = 0; i < nparam; i++)
      {
	old=analyseur; match_comma(); *newfun++ = entry();
	if (*analyseur == '=')
	{ 
	  analyseur++; argdef[i] = gclone(expr());
	}
	else argdef[i] = gclone(gzero);
      }
      argdef[i] = NULL; 
      analyseur+=2; /* match(')') && match('='); */

	/* record text */
      strncpy((char *)newfun, analyseur, len);
      ((char *) newfun)[len] = 0;

      ep->valence = EpUSER | (nparam << PARAMSHIFT);
      check_new_fun=NULL; analyseur = ch2;
      analyseurtetpil=avma; return gnil;
    }

    default:
      err(valencer1); return NULL; /* not reached */
  }
}

static long
number(long *nb)
{
  int m=0;
  for (*nb = 0; (*nb < 9) && isdigit(*analyseur); (*nb)++)
    m = 10*m + (*analyseur++ - '0');
  return m;
}

static GEN
constante()
{
  static long pw10[] = { 1, 10, 100, 1000, 10000, 100000, 1000000,
                        10000000, 100000000, 1000000000 };
  long l,m=0,n=0,plus=1,nb, av = avma, limite = (avma+bot)>>1;
  GEN z,y;

  y = stoi(number(&nb));
  while (isdigit(*analyseur))
  {
    m = number(&nb); y = mulsi(pw10[nb], y);
    y = addsi(m, y);
    if (avma < limite)
      y = gerepileupto(av,y);
  }
  if ( *analyseur!='.' && *analyseur!='e' && *analyseur!='E' ) return y;
  if (*analyseur=='.')
  {
    analyseur++;
    while (isdigit(*analyseur))
    {
      m = number(&nb); y = mulsi(pw10[nb], y);
      y = addsi(m, y);
      if (avma < limite)
	y = gerepileupto(av,y);
      n -= nb;
    }
  }
  l=lgef(y); if (l<prec) l=prec;
  z=cgetr(l); affir(y,z);
  if (*analyseur == 'e' || *analyseur == 'E')
  {
    char *old=analyseur;
    switch(*++analyseur)
    {
      case '-': plus=0; /* fall through */
      case '+': analyseur++;
    }
    m = number(&nb);
    if (isdigit(*analyseur)) err(expter1,old,mark.start);
    if (plus) n += m; else n -= m;
  }
  if (n)
  {
    y=cgetr(l); affsr(10,y); y = gpuigs(y, labs(n));
    z = n > 0 ?  mulrr(z,y) : divrr(z,y);
  }
  return z;
}

/* Looking for entry in hashtable table. n is hashing value */
entree *
findentry(long n, char *beg, char *end, entree **table)
{
  char *u, *v;
  entree *ep;

  if (!table) return NULL;
  for (ep = table[n]; ep; ep = ep->next)
  {
    u=ep->name; v=beg; while (*u && *u == *v) { u++; v++; }
    if (!*u && (v == end)) return ep;
  }
  if (foreignAutoload)
  { /* Try to autoload. */
    if ((*foreignAutoload)(beg, end - beg))
    {
      for (ep = table[n]; ep; ep = ep->next)
      {
	u=ep->name; v=beg; while (*u && *u == *v) { u++; v++; }
	if (!*u && (v == end)) return ep;
      }
      err(talker,"foreignAutoload reported success, but installed nothing");
    }
  }
  return NULL; /* not found */
}

static entree *
installep(void *f, char *name, int len, int valence, int add, int pos, entree **table)
{
  entree *ep = (entree *) xmalloc(sizeof(entree) + add + len + 1);
  const entree *ep1 = ep+1;
  char *u = (char *) ep1 + add;

  ep->name    = u; strncpy(u, name,len); u[len]=0;
  ep->args    = NULL; ep->help = NULL; ep->code = NULL;
  ep->value   = (void *) (f ? f : ep1);
  ep->next    = table[pos];
  ep->valence = valence;
  ep->menu    = 0;
  return table[pos] = ep;
}

long
manage_var(long n, entree *ep)
{
  static long max_avail = MAXVARN; /* first user variable not yet used */
  static long nvar; /* first GP free variable */
  long var;
  GEN p;

  if (n) /* special behaviour */
  {
    switch(n)
    {
      case 2: return nvar=0;
      case 3: return nvar;
    }

    /* user wants to delete one of his variables */
    if (max_avail == MAXVARN-1) return 0; /* nothing to delete */
    max_avail++;
    free(polx[max_avail]); /* frees both polun and polx */
    return max_avail+1;
  }

  if (nvar == max_avail) err(trucer1,mark.identifier,mark.start);
  if (ep)
  {
    p = (GEN)ep->value;
    var=nvar++;
  }
  else
  {
    p = (GEN) xmalloc(7*sizeof(long));
    var=max_avail--;
  }

  /* create polx[var] */
  p[0] = evaltyp(t_POL) | evallg(4);
  p[1] = evalsigne(1) | evallgef(4) | evalvarn(var);
  p[2] = zero; p[3] = un;
  polx[var] = p;

  /* create polun[nvar] */
  p += 4;
  p[0] = evaltyp(t_POL) | evallg(3);
  p[1] = evalsigne(1) | evallgef(3) | evalvarn(var);
  p[2] = un;
  polun[var] = p;

  varentries[var] = ep;
  if (ep) { polvar[nvar] = (long) ep->value; setlg(polvar, nvar+1); }
  return var;
}

long
fetch_var()
{
  return manage_var(0,NULL);
}

long
delete_var()
{
  return manage_var(1,NULL);
}

/* Find entry or create it */
static entree *
entry(void)
{
  char *old = analyseur;
  const long hash = hashvalue(&analyseur,functions_tblsz,1);
  entree *ep = findentry(hash,old,analyseur,functions_hash);
  long val,n;

  if (ep) return ep;
  if (compatible == WARN)
  {
    ep = findentry(hash,old,analyseur,funct_old_hash);
    if (ep) return ep; /* the warning was done in skipentry() */
  }

  /* ep does not exist. Create it */
  if (*analyseur == '(') 
    { n=0; val=EpNEW; } 
  else
    { n=7*sizeof(long); val=EpVAR; }
  ep = installep(NULL,old,analyseur-old,val,n,hash,functions_hash);

  if (n) manage_var(0,ep); /* Variable */
  return ep;
}

/********************************************************************/
/**                                                                **/
/**                          SKIP FUNCTIONS                        **/
/**                                                                **/
/********************************************************************/

/* [Ptitboul] Comme skipseq, sans modifier analyseur et autres, */
/* pour tre appel de l'extrieur.                             */
static char*
doskipseq(char *c)
{
  char *olds = analyseur;

  mark.start = c; analyseur = c; skipseq();
  c = analyseur; analyseur = olds;
  return c;
}

static void
skipstring()
{
  match('"');
  while (*analyseur)
    switch (*analyseur++)
    {
      case '"': return;
      case '\\': analyseur++;
    }
  match('"');
}

static void
skip_matrix_block(int no_affect)
{
  while (*analyseur == '[')
  {
    analyseur++;
    if (*analyseur == ',') { analyseur++; skipexpr(); }
    else
    {
      skipexpr();
      if (*analyseur == ',')
	if (*++analyseur != ']') skipexpr();
    }
    match(']');
  }

  if (*analyseur == '=' && analyseur[1] != '=')
  {
    if (no_affect) err(caracer1,analyseur,mark.start);
    analyseur++; skipexpr(); return;
  }
  if (can_repeat_op(*analyseur) && (*analyseur == analyseur[1]))
  {
    if (no_affect) err(caracer1,analyseur,mark.start);
    analyseur+=2; return;
  }
  if (!*analyseur) return;
  if (analyseur[1] != '=')
  {
    switch(*analyseur)
    {
      case '>': case '<':
	if (analyseur[1] != *analyseur || analyseur[2] != '=') return;
	if (no_affect) err(caracer1,analyseur,mark.start);
	analyseur+=3; skipexpr(); return;
      case '\\':
	if (analyseur[1] != '/' || analyseur[2] != '=') return;
	if (no_affect) err(caracer1,analyseur,mark.start);
	analyseur+=3; skipexpr(); return;
    }
    return;
  }
  
  switch(*analyseur)
  {
    case '+': case '-': case '*': case '/': case '\\': case '%':
      if (no_affect) err(caracer1,analyseur,mark.start);
      analyseur+=2; skipexpr(); return;
  }
}

static void
skipseq(void)
{
  for(;;)
  {
    while (separe(*analyseur)) analyseur++;
    if (!*analyseur || *analyseur == ')' || *analyseur == ',') return;
    skipexpr(); if (!separe(*analyseur)) return;
  }
}

static void
skipexpr(void)
{
  long niveau=3,e1,e2,e3;

  e1=e2=e3=0;
  for(;;)
    switch(niveau)
    {
      case 3:
	e3=1; skipfacteur();
	switch(*analyseur)
	{
	  case '*': case '/': case '%':
	    analyseur++; break;
	  case '\\':
	    if (*++analyseur == '/') analyseur++; break;
	  case '<': case '>':
	    if (analyseur[1]==*analyseur) { analyseur +=2; break; }
	    /* fall through */
	  default: niveau--;
	}
	break;

      case 2:
	if (!e3) { niveau++; break; }
	e3=0; e2=1;
	switch(*analyseur)
	{
	  case '+': case '-':
	    analyseur++; niveau++; break;
	  default: niveau--;
	}
	break;

      case 1:
	if (!e2) { niveau++; break; }
	e2=0; e1=1;
	switch(*analyseur)
	{
	  case '<':
	    switch(*++analyseur)
	    {
	      case '=': case '>': analyseur++;
	      default : niveau++;
	    }
	    break;
	  case '>':
	    if (*++analyseur == '=') analyseur++;
	    niveau++; break;
	  case '=': case '!':
	    if (analyseur[1] == '=') { analyseur+=2; niveau++; }
	    break;
	  default: niveau--;
	}
	break;

      case 0:
	if (!e1) { niveau++; break; }
	e1=0;
	switch(*analyseur)
	{
	  case '&': if (*++analyseur == '&') analyseur++; niveau++; break;
	  case '|': if (*++analyseur == '|') analyseur++; niveau++; break;
	  default: return;
	}
    }
}

static void
skipfacteur(void)
{
  if (*analyseur == '+' || *analyseur == '-') analyseur++;
  skiptruc();
  for(;;)
    switch(*analyseur)
    {
      case '.':
	analyseur++; while (isalnum(*analyseur)) analyseur++; break;
      case '^':
	analyseur++; skipfacteur(); break;
      case '~': case '_': case '\'':
	analyseur++; break;
      case '[':
	skip_matrix_block(1); break;
      case '!':
	if (analyseur[1] != '=') { analyseur++; break; }
      default: return;
    }
}

static void
skiptruc(void)
{
  long n=0;
  char *old;

  if (*analyseur == '!') { analyseur++; skiptruc(); return; }
  if (isalpha(*analyseur)) { skipidentifier(); return; }
  if (isdigit(*analyseur) || (*analyseur== '.')) { skipconstante(); return; }
  switch(*analyseur++)
  {
    case '(':
      skipexpr(); match(')'); return;
    case '[':
      old = analyseur-1;
      if (*analyseur == ';' && analyseur[1] == ']') { analyseur+=2; return; }
      if (*analyseur != ']')
      {
	do { n++; skipexpr(); old=analyseur; }
	while (*analyseur++ == ',');
	analyseur--;
      }
      switch (*analyseur)
      {
	case ']': analyseur++; return;
	case ';':
	{
	  long m,norig=n;
	  old=analyseur;
	  do {
	    m=n;
	    do { n++; analyseur++; skipexpr(); }
	    while (*analyseur != ';' && *analyseur != ']');
	    if (n-m != norig) err(recter1,old,mark.start);
	  } while (*analyseur != ']');
	  analyseur++; return;
	 }
	default:
	  err(vectmater1,old,mark.start);
	  return; /* not reached */
      }
    case '%':
      if (*analyseur == '`') { while (*++analyseur == '`'); return; }
      number(&n); return;
  }
  err(caracer1,analyseur-1,mark.start);
}

static void
skipidentifier(void)
{
  long matchcomma=0;
  entree *ep;
  char *old;

  mark.identifier = analyseur; ep = do_alias(skipentry());
  if (ep->code)
  {
    char *s = ep->code;

    if (*analyseur != '(')
    {
      if (EpVALENCE(ep) == 0) return; /* no mandatory argument */
      match('('); /* ==> error */
    }
    analyseur++;

    /* Optimized for G and p. */
    while (*s == 'G') { match_comma(); skipexpr(); s++; }
    if (*s == 'p') s++;
    while (*s) switch (*s++)
    {
      case 'G': case 'n': case 'L':
	  match_comma();
	  if (*analyseur == ',' || *analyseur == ')') break;
	  skipexpr(); break;
      case 'I':
	  match_comma(); skipseq(); break;
      case 'r':
      {
	  match_comma();
	  while (*analyseur)
	  {
	    while (*analyseur == '"') skipstring();
	    if (*analyseur == ',' || *analyseur == ')') break;
	    analyseur++;
	  }
	  break;
      }
      case 's':
      {
	  match_comma();
	  if (*s == '*')
	  {
	    while (*analyseur)
	    {
	      while (*analyseur == '"') skipstring();
	      if (*analyseur == ')') break;
	      if (*analyseur == ',') analyseur++;
	      else skipexpr();
	    }
	    s++; if (*s == 'p' || *s == 't') s++;
	    break;
	  }

	  while (*analyseur)
	  {
	    while (*analyseur == '"') skipstring();
	    if (*analyseur == ')' || *analyseur == ',') break;
	    skipexpr();
	  }
	  break;
      }
      case 'V': case 'S':
	  match_comma(); old = analyseur;
	  if (!isalpha(*analyseur)) err(varer1,old,mark.start);
	  if (EpVALENCE(skipentry()) != EpVAR && *(s-1) == 'V')
	    err(varer1,old,mark.start);
	  break;

      case 'p': case 'P': case 'l': case 'v': case 'f': case 'F': case 'x':
	  break;

      case 'D':
	  if ( *analyseur == ')' ) { analyseur++; return; }
	  while (*s++ != ',');
	  break;
      case '=':
	  match('='); matchcomma = 0; break;
      case ',':
          matchcomma=1; break;
      default:
	  err(talker,"Unknown type designator during skip, please report");
    }
    match(')');
    return;
  }
  if (EpVALENCE(ep) < EpUSER) /* fonctions predefinies */
  {
    if (!EpVALENCE(ep) && (*analyseur != '(')) return;
    match('(');
    switch(EpVALENCE(ep))
    {
      case 61: skipexpr(); break;
      case 60:
      {
	long m;
	old=analyseur; m=readlong();
	if ( m>=100 || m<0 ) err(labeler,old,mark.start);
	labellist[m]=analyseur; break;
      }
      case 50: skiptruc();
	if (*analyseur == '^') { analyseur++; skipfacteur(); };
	break;
      case 80: skipexpr(); match(','); skipseq();
          if (*analyseur != ')') { match(','); skipseq(); }
	  break;
      case 81: case 82: skipexpr(); match(','); skipseq(); break;
      default: err(valencer1);
    }
    match(')'); return;
  }
  switch (EpVALENCE(ep))
  {
    case EpVAR: /* variables */
      skip_matrix_block(0); return;

    case EpUSER: /* fonctions utilisateur */
    {
      char *ch1 = analyseur;
      GEN *arg = ep->args;

      if (*analyseur != '(')
      {
	if ( *analyseur != '='  ||  analyseur[1] == '=' ) return;
	match('('); /* error */
      }
      analyseur++;  /* skip '(' */
      while (*arg++)
      {
	if (do_switch(0,matchcomma)) matchcomma=1;
	else { match_comma(); skipexpr(); }
      }

      if (*analyseur == ')')
	if ( analyseur[1] != '=' || analyseur[2] == '=' )
	  { analyseur++; return; }

      /* here we are redefining a user function */
      old = analyseur;
      if (*analyseur != ',' && *analyseur != ')') skipexpr();
      while (*analyseur == ',') { analyseur++; skipexpr(); }
      match(')');

      if (*analyseur != '=' || analyseur[1] == '=')
        if (skipping_fun_def) return;
	else
	  err(nparamer1,old,mark.start);
      analyseur = ch1; matchcomma=0;
    } /* fall through */

    case EpNEW: /* nouvelle fonction */
      if (check_new_fun && ! skipping_fun_def)
      {
	err_new_fun(); /* ep not created yet: no need to kill it */
	err(paramer1, mark.identifier, mark.start);
      }
      check_new_fun = NOT_CREATED_YET; match('(');
      while (*analyseur != ')')
      { 
	match_comma(); skipexpr();
      };
      match(')');
      if (*analyseur == '=')
      { 
	skipping_fun_def++;
	analyseur++; skipseq();
	skipping_fun_def--;
      }
      check_new_fun=NULL; return;

    default: err(valencer1);
  }
}

static void
skipconstante(void)
{
  while (isdigit(*analyseur)) analyseur++;
  if ( *analyseur!='.' && *analyseur!='e' && *analyseur!='E' ) return;
  if (*analyseur=='.') analyseur++;
  while (isdigit(*analyseur)) analyseur++;
  if ( *analyseur=='e'  ||  *analyseur=='E' )
  {
    analyseur++;
    if ( *analyseur=='+' || *analyseur=='-' ) analyseur++;
    while (isdigit(*analyseur)) analyseur++;
  }
}

static entree *
skipentry(void)
{
  static entree fakeEpNEW = { "",EpNEW,0,0,0,NULL,NULL,NULL };
  static entree fakeEpVAR = { "",EpVAR,0,0,0,NULL,NULL,NULL };
  char *old=analyseur;
  const long hash=hashvalue(&analyseur,functions_tblsz,1);
  entree *ep = findentry(hash, old, analyseur, functions_hash);

  if (ep) return ep;
  if (compatible == WARN)
  {
    ep = findentry(hash,old,analyseur,funct_old_hash);
    if (ep)
    {
      fprintferr("  *** Warning: using obsolete function %s\n",ep->name);
      return ep;
    }
  }
  return (*analyseur == '(') ? &fakeEpNEW : &fakeEpVAR;
}

/********************************************************************/
/**                                                                **/
/**                          MEMBER FUNCTIONS                      **/
/**                                                                **/
/********************************************************************/

static GEN
bnf(GEN bnr)
/* get bnf. Here bnr can be a bnrinit (with or without gen) or a bnfinit
 (with or without fu) */
{
  if (typ(bnr) == t_VEC)
    switch(lg(bnr))
    {
      case 11: return bnr; /* bnfinit */
      case 7: return (GEN) bnr[1]; /* bnrinit */
    }

  err(member,"bnf",mark.member,mark.start);
  return NULL; /* not reached */
}

static GEN
nf(GEN bnf)
/* get nf. Here bnf can be a bnfinit (with or without fu), a bnrinit (with
   or without gen) or a nfinit. */
{
  if (typ(bnf) == t_VEC)
    switch(lg(bnf))
    {
      case 11: return (GEN) bnf[7]; /* bnfinit */
      case 10: return bnf; /* nfinit */
      case 7: return gmael(bnf,1,7); /* bnrinit */
    }

  err(member,"nf",mark.member,mark.start);
  return NULL; /* not reached */
}

static GEN
zk(GEN bnf)
/* get integral basis. Here bnf can be a bnfinit (with or without fu), a
   bnfclass (with or without fu), a bnrinit (with or without gen) or a nfinit.
   */
{
  switch(typ(bnf))
  {
    case t_MAT:
      /* bnfclass */
      bnf=(GEN)bnf[1]; if (lg(bnf)<8) break;
      return (GEN) bnf[4];
    case t_VEC:
      switch(lg(bnf))
      {
	case 11:
	  return gmael(bnf,7,7);
	case 10:
	  return (GEN) bnf[7];
	case 7:
	  return gmael3(bnf,1,7,7);
      }
  }
  err(member,"bnf",mark.member,mark.start);
  return NULL; /* not reached */
}

static GEN
disc(GEN bnf)
/* get discriminant. Here bnf can be a bnfinit (with or without fu), a
   bnfclass (with or without fu), a bnrinit (with or without gen) or a nfinit.
   */
{
  switch(typ(bnf))
  {
    case t_MAT:
      /* bnfclass */
      bnf=(GEN)bnf[1]; if (lg(bnf)<8) break;
      return gmael(bnf,3,1);
    case t_VEC:
      switch(lg(bnf))
      {
	case 14: case 20:
	  return (GEN) bnf[12]; /* elliptic curve */
	case 11:
	  return gmael(bnf,7,3);
	case 10:
	  return (GEN) bnf[3];
	case 7:
	  return gmael3(bnf,1,7,3);
      }
  }
  err(member,"disc",mark.member,mark.start);
  return NULL; /* not reached */
}

static GEN
pol(GEN bnf)
/* get polynomial. Here bnf can be a bnfinit (with or without fu), a
   bnrinit (with or without gen) a nfinit, a pol or a quad. */
{
  if (typ(bnf)==t_VEC)
    switch(lg(bnf))
    {
      case 11: return gmael(bnf,7,1);
      case 10: return (GEN) bnf[1];
      case 7: return gmael3(bnf,1,7,1);
    }
  else if (typ(bnf)==t_POL) return bnf;
  else if (typ(bnf)==t_QUAD) return (GEN) bnf[2];
  err(member,"pol",mark.member,mark.start);
  return NULL; /* not reached */
}

static GEN
mod(GEN x)
/* get modulus. Here x can be an intmod, a polmod or a quad */
{
  if ((typ(x)==t_INTMOD)||(typ(x)==t_POLMOD)) return (GEN) x[1];
  else if (typ(x)==t_QUAD) return (GEN) x[2];
  err(member,"mod",mark.member,mark.start);
  return NULL; /* not reached */
}

static GEN
sign(GEN bnf)
/* get signature. Here bnf can be a bnfinit (with or without fu), a
   bnrinit (with or without gen) or a nfinit. */
{
  if (typ(bnf)==t_VEC)
    switch(lg(bnf))
    {
      case 11: return gmael(bnf,7,2);
      case 10: return (GEN) bnf[2];
      case 7: return gmael3(bnf,1,7,2);
    }
  err(member,"sign",mark.member,mark.start);
  return NULL; /* not reached */
}

static GEN
t2(GEN bnf)
/* get t2 matrix. Here bnf can be a bnfinit (with or without fu), a
   bnrinit (with or without gen) or a nfinit. */
{
  if (typ(bnf)==t_VEC)
    switch(lg(bnf))
    {
      case 11: return gmael3(bnf,7,5,3);
      case 10: return gmael(bnf,5,3);
      case 7: return gmael4(bnf,1,7,5,3);
    }
  err(member,"t2",mark.member,mark.start);
  return NULL; /* not reached */
}

static GEN
diff(GEN bnf)
/* get different. Here bnf can be a bnfinit (with or without fu), a
   bnrinit (with or without gen) or a nfinit. */
{
  if (typ(bnf)==t_VEC)
    switch(lg(bnf))
    {
      case 11: return gmael3(bnf,7,5,5);
      case 10: return gmael(bnf,5,5);
      case 7: return gmael4(bnf,1,7,5,5);
    }
  err(member,"diff",mark.member,mark.start);
  return NULL; /* not reached */
}

static GEN
codiff(GEN bnf)
/* get codifferent. Here bnf can be a bnfinit (with or without fu), a
   bnrinit (with or without gen) or a nfinit. */
{
  if (typ(bnf)==t_VEC)
    switch(lg(bnf))
    {
      case 11: return gdiv(gmael3(bnf,7,5,7), gmael(bnf,7,3));
      case 10: return gdiv(gmael(bnf,5,7),(GEN) bnf[3]);
      case 7: return gdiv(gmael4(bnf,1,7,5,7), gmael3(bnf,1,7,3));
    }
  err(member,"codiff",mark.member,mark.start);
  return NULL; /* not reached */
}

static GEN
mroots(GEN bnf)
/* get roots. Here bnf can be a bnfinit (with or without fu), a
   bnrinit (with or without gen) or a nfinit. */
{
  if (typ(bnf)==t_VEC)
    switch(lg(bnf))
    {
      case 14: case 20: return (GEN)bnf[14]; /* elliptic curve */
      case 11: return gmael(bnf,7,6);
      case 10: return (GEN) bnf[6];
      case 7: return gmael3(bnf,1,7,6);
    }
  err(member,"roots",mark.member,mark.start);
  return NULL; /* not reached */
}

static GEN
clgp(GEN bnf)
/* get class group components (3-component row vector); bnf can be a
   bnfinit (with or without fu), a bnfclass (with or without fu), or a
   bnrinit (with or without gen) */
{
  long i;
  GEN p1;

  switch(typ(bnf))
  {
    case t_MAT:
      /* bnfclass */
      bnf=(GEN)bnf[1]; if (lg(bnf)<8) break;
      return (GEN) bnf[5];
    case t_VEC:
      /* bnfinit or bnrinit */
      switch(lg(bnf))
      {
	case 11:
	  bnf = (GEN) bnf[8]; if (lg(bnf)<5) break;
	  return (GEN) bnf[1];
	case 7:
	  return (GEN) bnf[5];
	case 5: case 6:
	  p1=cgetg(4,t_VEC);
	  for (i=1; i<=3; i++) p1[i] = bnf[i];
	  return p1; /* quadclassgroup */
	case 4:
	  return bnf;
      }
  }
  err(member,"clgp",mark.member,mark.start);
  return NULL; /* not reached */
}

static GEN
reg(GEN bnf)
/* get regulator; bnf can be a bnfinit (with or without fu), a bnfclass (with
   or without fu), or a quadclassgroup */
{
  switch(typ(bnf))
  {
    case t_MAT:
      /* bnfclass */
      bnf=(GEN)bnf[1]; if (lg(bnf)<8) break;
      return (GEN) bnf[6];
    case t_VEC:
      /* bnfinit or bnrinit */
      switch(lg(bnf))
      {
	case 11:
	  bnf=(GEN)bnf[8]; if (lg(bnf)<5) break;
	  return (GEN) bnf[2];
	case 7:
	  err(impl,"ray regulator");
	  return NULL; /* not reached */
	case 6:
	  return (GEN)bnf[4]; /* quadclassgroup (real) */
	case 5:
	  return gun; /* quadclassgroup (imag) */
      }
  }
  err(member,"reg",mark.member,mark.start);
  return NULL; /* not reached */
}

static GEN
fu(GEN bnf)
/* get fundamental units; bnf can be a bnfinit (with or without fu),
   a bnfclass (with fu), or a quadclassgroup */
{
  GEN pol;

  switch(typ(bnf))
  {
    case t_MAT:
      /* bnfclass */
      bnf=(GEN)bnf[1]; if (lg(bnf)<11) break;
      return gmodulcp((GEN)bnf[9],(GEN)bnf[1]);
    case t_VEC:
      /* bnfinit or bnrinit */
      switch(lg(bnf))
      {
	case 11:
	  pol=gmael(bnf,7,1); bnf=(GEN)bnf[8];
	  switch(lg(bnf))
	  {
	    case 5:
	      bnf=buchfu(bnf); return gmodulcp((GEN)bnf[1],pol);
	    case 7:
	      return gmodulcp((GEN)bnf[5],pol);
	  }
	  break;
	case 7:
	  err(impl,"ray units");
	  return NULL; /* not reached */
	case 6:
	  bnf=discsr(gmael(bnf,3,1));
	  return fundunit(bnf); /* quadclassgroup (real) */
	case 5:
	  return cgetg(1,t_VEC); /* quadclassgroup (imag) */
      }
  }
  err(member,"fu",mark.member,mark.start);
  return NULL; /* not reached */
}

static GEN
tu(GEN bnf)
/* get torsion units; bnf can be a bnfinit (with or without fu),
   a bnfclass (with fu), or a quadclassgroup. Result is a 2 element row
   vector [w,e] where w is the number of roots of 1, and e a generator as
   a polymod. */
{
  GEN pol,y;

  switch(typ(bnf))
  {
    case t_MAT:
      /* bnfclass */
      bnf = (GEN) bnf[1]; if (lg(bnf)<11) break;
      y=cgetg(3,t_VEC); y[1]=mael(bnf,8,1);
      y[2]=lmodulcp(gmael(bnf,8,2),(GEN) bnf[1]);
      return y;
    case t_VEC:
      /* bnfinit or bnrinit */
      switch(lg(bnf))
      {
	case 11:
	  pol=gmael(bnf,7,1); bnf=(GEN)bnf[8];
	  if (lg(bnf)<5) break;
	  y=cgetg(3,t_VEC); y[1]=mael(bnf,4,1);
	  y[2]=lmodulcp(gmael(bnf,4,2),pol); return y;
	case 7:
	  err(impl,"ray torsion units");
	  return NULL; /* not reached */
	case 6:
	  y=cgetg(3,t_VEC); y[1]=deux; y[2]=lstoi(-1); return y;
	   /* quadclassgroup (real) */
      }
  }
  err(member,"tu",mark.member,mark.start);
  return NULL; /* not reached */
}

static GEN
futu(GEN bnf)
/*  concatenation of fu and tu, w is lost */
{
  GEN fuc = fu(bnf), tuc = (GEN) tu(bnf)[2];
  return concat(fuc,tuc);
}

static GEN
tufu(GEN bnf)
/*  concatenation of tu and fu, w is lost */
{
  GEN fuc = fu(bnf), tuc = (GEN) tu(bnf)[2];
  return concat(tuc,fuc);
}

static GEN
zkst(GEN bid)
/* structure of (Z_K/m)^*, where bid is a idealstarinit (with or without gen)
   or a bnrinit (with or without gen) */
{
  if (typ(bid)==t_VEC)
    switch(lg(bid))
    {
      case 6: return (GEN) bid[2];   /* idealstarinit */
      case 7: return gmael(bid,2,2); /* bnrinit */
    }
  err(member,"zkst",mark.member,mark.start);
  return NULL; /* not reached */
}

static GEN
no(GEN clgp)
/* number of elements of a group */
{
  if ( typ(clgp)!=t_VEC  || (lg(clgp)!=3 && lg(clgp)!=4) )
    err(member,"no",mark.member,mark.start);
  return (GEN) clgp[1];
}

static GEN
cyc(GEN clgp)
/* cyclic decomposition (SNF) of a group */
{
  if ( typ(clgp)!=t_VEC  || (lg(clgp)!=3 && lg(clgp)!=4) )
    err(member,"cyc",mark.member,mark.start);
  return (GEN) clgp[2];
}

static GEN
gen(GEN clgp)
/* SNF generators of a group */
{
  if (typ(clgp)!=t_VEC || lg(clgp)!=4)
    err(member,"gen",mark.member,mark.start);
  return (GEN) clgp[3];
}

#define is_ell(x) (typ(x) == t_VEC && lg(x)>=14)
#define is_bigell(x) (typ(x) == t_VEC && lg(x)>=20)

static GEN
a1(GEN x)
{
  if (!is_ell(x)) err(member,"a1",mark.member,mark.start);
  return (GEN)x[1];
}

static GEN
a2(GEN x)
{
  if (!is_ell(x)) err(member,"a2",mark.member,mark.start);
  return (GEN)x[2];
}

static GEN
a3(GEN x)
{
  if (!is_ell(x)) err(member,"a3",mark.member,mark.start);
  return (GEN)x[3];
}

static GEN
a4(GEN x)
{
  if (!is_ell(x)) err(member,"a4",mark.member,mark.start);
  return (GEN)x[4];
}

static GEN
a6(GEN x)
{
  if (!is_ell(x)) err(member,"a6",mark.member,mark.start);
  return (GEN)x[5];
}

static GEN
b2(GEN x)
{
  if (!is_ell(x)) err(member,"b2",mark.member,mark.start);
  return (GEN)x[6];
}

static GEN
b4(GEN x)
{
  if (!is_ell(x)) err(member,"b4",mark.member,mark.start);
  return (GEN)x[7];
}

static GEN
b6(GEN x)
{
  if (!is_ell(x)) err(member,"b6",mark.member,mark.start);
  return (GEN)x[8];
}

static GEN
b8(GEN x)
{
  if (!is_ell(x)) err(member,"b8",mark.member,mark.start);
  return (GEN)x[9];
}

static GEN
c4(GEN x)
{
  if (!is_ell(x)) err(member,"c4",mark.member,mark.start);
  return (GEN)x[10];
}

static GEN
c6(GEN x)
{
  if (!is_ell(x)) err(member,"c6",mark.member,mark.start);
  return (GEN)x[11];
}

static GEN
j(GEN x)
{
  if (!is_ell(x)) err(member,"j",mark.member,mark.start);
  return (GEN)x[13];
}

static GEN
momega(GEN x)
{
  GEN y;

  if (!is_bigell(x)) err(member,"omega",mark.member,mark.start);
  y=cgetg(3,t_VEC); y[1]=x[15]; y[2]=x[16];
  return y;
}

static GEN
meta(GEN x)
{
  GEN y;

  if (!is_bigell(x)) err(member,"eta",mark.member,mark.start);
  y=cgetg(3,t_VEC); y[1]=x[17]; y[2]=x[18];
  return y;
}

static GEN
area(GEN x)
{
  if (!is_bigell(x)) err(member,"area",mark.member,mark.start);
  return (GEN)x[19];
}

/*
 * Only letters and digits in member names. AT MOST 8 of THEM
 * (or modify gp_rl.c::pari_completion)
 */
default_type gp_member_list[] =
{
  { "a1",(void*)a1 },
  { "a2",(void*)a2 },
  { "a3",(void*)a3 },
  { "a4",(void*)a4 },
  { "a6",(void*)a6 },
  { "area",(void*)area },
  { "b2",(void*)b2 },
  { "b4",(void*)b4 },
  { "b6",(void*)b6 },
  { "b8",(void*)b8 },
  { "bnf",(void*)bnf },
  { "c4",(void*)c4 },
  { "c6",(void*)c6 },
  { "clgp",(void*)clgp },
  { "codiff",(void*)codiff },
  { "cyc",(void*)cyc },
  { "diff",(void*)diff },
  { "disc",(void*)disc },
  { "eta",(void*)meta },
  { "fu",(void*)fu },
  { "futu",(void*)futu },
  { "gen",(void*)gen },
  { "j",(void*)j },
  { "mod",(void*)mod },
  { "nf",(void*)nf },
  { "no",(void*)no },
  { "omega",(void*)momega },
  { "pol",(void*)pol },
  { "reg",(void*)reg },
  { "roots",(void*)mroots },
  { "sign",(void*)sign },
  { "t2",(void*)t2 },
  { "tu",(void*)tu },
  { "tufu",(void*)tufu },
  { "zk",(void*)zk },
  { "zkst",(void*)zkst },

  { NULL,NULL }
};

static GEN
read_member(GEN x)
{
  static char buf[10];
  default_type dflt;
  long n;

  mark.member = analyseur;
  for (n=0; isalnum(*analyseur) && n<9; n++)
    buf[n] = *analyseur++;
  buf[n]=0; n=0;
  do
  {
    dflt=gp_member_list[n++];
    if (!strcmp(buf,dflt.name))
      return ((GEN (*)(ANYARG)) dflt.fun)(x);
  } while (gp_member_list[n].name);

  err(talker2,"Unknown member function : ",mark.member,mark.start);
  return NULL; /* not reached */
}

/********************************************************************/
/**                                                                **/
/**                        SIMPLE GP FUNCTIONS                     **/
/**                                                                **/
/********************************************************************/

long
check_break_status(long action, GEN *x)
{
  static long status, count;
  static GEN res;

  if (action == NOACTION)
  {
    if (x) *x = res;
    return status;
  }

  switch(action)
  {
    case DOCLEAR:
      status = br_NONE; res = gnil; return 0;
    case DOBREAK:
      status = br_BREAK; count = (long) x; return 1;
    case DONEXT:
      status = br_NEXT; return 1;
    case DORETURN:
      status = br_RETURN; res = *x; return 1;

    case DOLOOP: /* for, while, sum */
      switch(status)
      {
	case br_NEXT:
	  status=br_NONE; /* fall through */
	case br_NONE:
	  return 0;

        case br_BREAK:
	  count--;
	  if (!count) status=br_NONE; /* fall through */
        case br_RETURN:
	  return 1;
      }

    case DOFUNCTION:
      switch(status)
      {
	case br_NEXT:
	  status=br_NONE; /* fall through */
	case br_NONE:
	  return 0;

        case br_BREAK: /* possibly "bad break" ?? */
	  status=br_NONE; *x = gnil; return 1;

        case br_RETURN:
	  status=br_NONE; *x = res; return 1;
      }
  }
  return 0; /* not reached */
}

static void
return0(GEN x)
{
  check_break_status(DORETURN, &x);
}

static void
next0()
{
  check_break_status(DONEXT, NULL);
}

static void
break0(long n)
{
  if (n<1)
    err(talker2,"I need a positive integer: ",mark.identifier,mark.start);
  check_break_status(DOBREAK, (GEN *)n);
}

static GEN
extract0(GEN x, GEN l1, GEN l2)
{
  if (gcmp_1(l2)) return extract(x,l1);
  return matextract(x,l1,l2);
}

static void
alias0(char *s, char *old)
{
  char *end=old;
  long hash = hashvalue(&end,functions_tblsz,1);
  entree *ep = findentry(hash,old,end,functions_hash);

  if (!ep) err(talker2,"Unknown function: ",mark.raw,mark.start);
  if (EpVALENCE(ep)==EpVAR)
    err(talker2,"Only functions can be aliased: ",mark.raw,mark.start);

  end=s; hash=hashvalue(&end,functions_tblsz,1);
  if (findentry(hash,s,end,functions_hash))
    err(talker2, "Can't replace an existing symbol by an alias:\n   ",
        mark.raw, mark.start);
  ep = do_alias(ep);
  installep((void *)ep, s, strlen(s), EpALIAS, 0, hash, functions_hash);
}

static GEN
gshiftplus(GEN x, GEN n)
{
  return gshift(x,itos(n));
}

static GEN
gshiftminus(GEN x, GEN n)
{
  return gshift(x,-itos(n));
}

GEN
geni(void)
{
  return gi;
}
